"""
HDF5-backed FX Rates

The classes defined in this file produce and consume foreign exchange rates
stored in HDF5 files.

An FX Rate file contains one or more "rates", each of which contains a
collection of arrays exchange rates describing how to convert from "base"
currencies to "quote" currencies.

The specific set of rates contained in a particular file is arbitrary and
user-defined. We infer no particular semantics from their names, other than
that they are distinct rates. Pipeline API users can select the rate that
should be used for a currency conversion via ``BoundColumn.fx`` method.

Examples of possible rate names might be things like "bid", "mid", and "ask",
or "london_close", "tokyo_close", "nyse_close".

HDF5 FX Rate File Schema
------------------------

An FX rate file has two kinds of nodes: **data nodes** and **index nodes**.

Data nodes are 2-dimensional arrays of numerical exchange rate values. Data
nodes are stored at paths of the form /data/{rate}/{quote_currency}/rates. The
path to a data node encodes two pieces of information: the "rate" (e.g. 'mid',
'bid', 'ask') of the node, and the quote currency of the node. All data nodes
in a file have the same shape.

Index nodes are 1-dimensional arrays containing row- and column-labels for the
file's data nodes. There are two index nodes in a file, at /index/dts and
/index/currencies.

**Example:**

A file containing fields ["rate1", "rate2"] and currencies ["USD", "CAD",
"EUR"] has the following schema:

.. code-block:: none

   /data
     /field1
       /USD
         /rates
       /CAD
         /rates
       /EUR
         /rates
     /field2
       /USD
         /rates
       /CAD
         /rates
       /EUR
         /rates
   /index
     /dts
     /currencies

/index/dts
^^^^^^^^^^

This node contains a 1-dimensional array of int64s. When interpreted as an
np.datetime64[ns], each entry in the array defines the date label for the
corresponding column of a data node.

/index/currencies
^^^^^^^^^^^^^^^^^

This node contains a 1-dimensional array of length-3 strings. Each entry in the
array defines the the label for the corresponding row of a data node.

/data/{rate}/{quote_currency}/rates
^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^

Nodes at these locations are data nodes. They store 2-dimensional arrays of
exchange rates.

Represented as a DataFrame with appropriate indices, the data at
/data/rate1/USD/rates might look like this::

        2014-01-01  2014-01-02  2014-01-03  2014-01-04  2014-01-05
   USD    1.000000    1.000000    1.000000    1.000000    1.000000
   EUR    0.808477    2.235770    0.378603    0.090983    1.135781
   CAD    0.442329    0.817352    1.181822    0.198182    0.536609

Each row of the array contains exchange rates mapping from currency of the
row's label to the quote encoded in the node's path (USD in this
example). The label for row i in a data node is the ith element of
/index/currencies.

Each column of the array contains exchange rates for a given date. The label
for column i in a data node is the ith element of /index/dts.
"""

import h5py
import logging
import numpy as np
import pandas as pd
import structlog

from ziplime.utils.numpy_utils import bytes_array_to_native_str_object_array

from .base import FXRateReader, DEFAULT_FX_RATE
from .utils import check_dts, is_sorted_ascending

HDF5_FX_VERSION = 0
HDF5_FX_DEFAULT_CHUNK_SIZE = 75

INDEX = "index"
DATA = "data"
CURRENCIES = "currencies"
DTS = "dts"
RATES = "rates"

log = structlog.get_logger(__name__)


class HDF5FXRateReader(FXRateReader):
    """An FXRateReader backed by HDF5.

    Parameters
    ----------
    group : h5py.Group
        Top-level group written by an :class:`HDF5FXRateWriter`.
    default_rate : str
        Rate to use when ``get_rates`` is called requesting the default rate.
    """

    def __init__(self, group, default_rate):
        self._group = group
        self._default_rate = default_rate

        if self.version != HDF5_FX_VERSION:
            raise ValueError(
                "FX Reader version ({}) != File Version ({})".format(
                    HDF5_FX_VERSION,
                    self.version,
                )
            )

    @classmethod
    def from_path(cls, path, default_rate):
        """Construct from a file path.

        Parameters
        ----------
        path : str
            Path to an HDF5 fx rates file.
        default_rate : str
            Rate to use when ``get_rates`` is called requesting the default
            rate.
        """
        return cls(h5py.File(path), default_rate=default_rate)

    @property
    def version(self):
        try:
            return self._group.attrs["version"]
        except KeyError:
            # TODO: Remove this.
            return 0

    @property
    def dts(self):
        """Column labels for rate groups."""
        raw_dts = self._group[INDEX][DTS][:].astype("M8[ns]")
        if not is_sorted_ascending(raw_dts):
            raise ValueError("dts are not sorted for {}!".format(self._group))

        return pd.DatetimeIndex(raw_dts, tz="UTC")

    @property
    def currencies(self):
        """Row labels for rate groups."""
        # Currencies are stored as fixed-length bytes in the file, but we want
        # `str` objects in memory.
        bytes_array = self._group[INDEX][CURRENCIES][:]
        objects = bytes_array_to_native_str_object_array(bytes_array)
        return pd.Index(objects)

    def get_rates(self, rate, quote, bases, dts):
        """Get rates to convert ``bases`` into ``quote``.

        See :class:`ziplime.data.fx.base.FXRateReader` for details.
        """
        if rate == DEFAULT_FX_RATE:
            rate = self._default_rate

        check_dts(dts)

        # TODO FIXME TZ MESS
        if dts.tzinfo is None:
            dts = dts.tz_localize(self.dts.tzinfo)
        col_ixs = self.dts.searchsorted(dts, side="right") - 1
        row_ixs = self.currencies.get_indexer(bases)

        try:
            dataset = self._group[DATA][rate][quote][RATES]
        except KeyError as exc:
            raise ValueError(
                "FX rates not available for rate={}, quote_currency={}.".format(
                    rate, quote
                )
            ) from exc

        # OPTIMIZATION: Column indices correspond to dates, which must be in
        # sorted order. Rather than reading the entire dataset from h5, we can
        # read just the interval from min_col to max_col inclusive
        #
        # However, we also need to handle two important edge cases:
        #
        #   1. row_ixs contains -1 for any currencies we don't know about.
        #   2. col_ixs contains -1 for dts before the start of self.dts.
        #
        # If either of the above cases obtains, we want to return NaN for the
        # corresponding output locations.

        # We handle each of these cases by reading raw data into a buffer with
        # one extra column and one extra row. When we then permute the raw data
        # into the correct order, any row or column indices with values of -1
        # will pull from the extra row/column, which will always contain NaN.

        slice_begin = max(col_ixs[0], 0)
        slice_end = max(col_ixs[-1], 0) + 1  # +1 to be inclusive of end date.

        # Allocate a buffer full of NaNs with one extra column and row. See
        # OPTIMIZATION notes above.
        buf = np.full(
            (len(self.currencies) + 1, slice_end - slice_begin + 1),
            np.nan,
        )

        buf[:-1, :-1] = dataset[:, slice_begin:slice_end]

        # Permute the rows into place, pulling from the empty NaN locations for
        # row and column indices of -1.
        out = buf[:, col_ixs - slice_begin][row_ixs]

        # Transpose everything to maintain dts as row labels, currencies as col
        # labels which is expected everywhere else.
        return out.transpose()


class HDF5FXRateWriter:
    """Writer class for HDF5 files consumed by HDF5FXRateReader."""

    def __init__(self, group, date_chunk_size=HDF5_FX_DEFAULT_CHUNK_SIZE):
        self._group = group
        self._date_chunk_size = date_chunk_size

    def write(self, dts, currencies, data):
        """Write data to the file.

        Parameters
        ----------
        dts : pd.DatetimeIndex
            Index of row labels for rates to be written.
        currencies : np.array[object]
            Array of column labels for rates to be written.
        data : iterator[(str, str, np.array[float64])]
            Iterator of (rate, quote_currency, array) tuples. Each array
            should be of shape ``(len(dts), len(currencies))``, and should
            contain a table of rates where each column is a timeseries of rates
            mapping its column label's currency to ``quote_currency``.
        """

        if len(currencies):
            chunks = (len(currencies), min(self._date_chunk_size, len(dts)))
        else:
            # h5py crashes if we provide chunks for empty data.
            chunks = None

        self._write_metadata()
        self._write_index_group(dts, currencies)
        self._write_data_group(dts, currencies, data, chunks)

    def _write_metadata(self):
        self._group.attrs["version"] = HDF5_FX_VERSION
        self._group.attrs["last_updated_utc"] = str(pd.Timestamp.utcnow())

    def _write_index_group(self, dts, currencies):
        """Write content of /index."""
        if not is_sorted_ascending(dts):
            raise ValueError("dts is not sorted")

        for c in currencies:
            if not isinstance(c, str) or len(c) != 3:
                raise ValueError("Invalid currency: {!r}".format(c))

        index_group = self._group.create_group(INDEX)

        self._log_writing(INDEX, DTS)
        index_group.create_dataset(DTS, data=dts.astype("int64"))

        self._log_writing(INDEX, CURRENCIES)
        index_group.create_dataset(CURRENCIES, data=currencies.astype("S3"))

    def _write_data_group(self, dts, currencies, data, chunks):
        """Write content of /data."""
        data_group = self._group.create_group(DATA)

        expected_shape = (len(dts), len(currencies))
        for rate, quote, array in data:
            if array.shape != expected_shape:
                raise ValueError(
                    "Unexpected shape for rate={}, quote={}."
                    "\nExpected shape: {}. Got {}.".format(
                        rate, quote, expected_shape, array.shape
                    )
                )

            self._log_writing(DATA, rate, quote)
            target = data_group.require_group("/".join((rate, quote)))

            # Transpose the rates array so that the hdf5 file holds arrays
            # with currencies as row labels and dates as column labels. This
            # helps with compression, as the *rows* (rather than the columns)
            # all have similar values, which lends itself to the HDF5 file's
            # C-contiguous storage.
            target.create_dataset(
                RATES,
                data=array.transpose(),
                chunks=chunks,
                compression="lzf",
                shuffle=True,
            )

    def _log_writing(self, *path):
        log.debug("Writing {}", "/".join(path))



import math

from ziplime.finance.constants import ROOT_SYMBOL_TO_ETA, DEFAULT_ETA

from ziplime.assets.entities.futures_contract import FuturesContract
from ziplime.finance.slippage.market_impact_base import MarketImpactBase
from ziplime.utils.dummy import DummyMapping


class VolatilityVolumeShare(MarketImpactBase):
    """Model slippage for futures contracts according to the following formula:

        new_price = price + (price * MI / 10000),

    where 'MI' is market impact, which is defined as:

        MI = eta * sigma * sqrt(psi)

    - ``eta`` is a constant which varies by root symbol.
    - ``sigma`` is 20-day annualized volatility.
    - ``psi`` is the volume traded in the given bar divided by 20-day ADV.

    Parameters
    ----------
    volume_limit : float
        Maximum percentage (as a decimal) of a bar's total volume that can be
        traded.
    eta : float or dict
        Constant used in the market impact formula. If given a float, the eta
        for all futures contracts is the same. If given a dictionary, it must
        map root symbols to the eta for contracts of that symbol.
    """

    NO_DATA_VOLATILITY_SLIPPAGE_IMPACT = 7.5 / 10000
    allowed_asset_types = (FuturesContract,)

    def __init__(self, volume_limit, eta=ROOT_SYMBOL_TO_ETA):
        super(VolatilityVolumeShare, self).__init__()
        self.volume_limit = volume_limit

        # If 'eta' is a constant, use a dummy mapping to treat it as a
        # dictionary that always returns the same value.
        # NOTE: This dictionary does not handle unknown root symbols, so it may
        # be worth revisiting this behavior.
        if isinstance(eta, (int, float)):
            self._eta = DummyMapping(float(eta))
        else:
            # Eta is a dictionary. If the user's dictionary does not provide a
            # value for a certain contract, fall back on the pre-defined eta
            # values per root symbol.
            self._eta = {**ROOT_SYMBOL_TO_ETA, **eta}

    def __repr__(self):
        if isinstance(self._eta, DummyMapping):
            # Eta is a constant, so extract it.
            eta = self._eta["dummy key"]
        else:
            eta = "<varies>"
        return "{class_name}(volume_limit={volume_limit}, eta={eta})".format(
            class_name=self.__class__.__name__,
            volume_limit=self.volume_limit,
            eta=eta,
        )

    def get_simulated_impact(
            self,
            order,
            current_price,
            current_volume,
            txn_volume,
            mean_volume,
            volatility,
    ):
        try:
            eta = self._eta[order.asset.root_symbol]
        except Exception:
            eta = DEFAULT_ETA

        psi = txn_volume / mean_volume

        market_impact = eta * volatility * math.sqrt(psi)

        # We divide by 10,000 because this model computes to basis points.
        # To convert from bps to % we need to divide by 100, then again to
        # convert from % to fraction.
        return (current_price * market_impact) / 10000

    def get_txn_volume(self, data, order):
        volume = data.current(order.asset, "volume")
        return volume * self.volume_limit



import datetime
from functools import reduce, lru_cache
from operator import mul
from typing import Any

import polars as pl
import structlog
from exchange_calendars import ExchangeCalendar

from ziplime.assets.domain.continuous_future import ContinuousFuture
from ziplime.assets.entities.asset import Asset
from ziplime.assets.entities.equity import Equity
from ziplime.constants.data_type import DataType
from ziplime.constants.period import Period
from ziplime.data.services.data_source import DataSource
from ziplime.utils.date_utils import period_to_timedelta


class DataBundle(DataSource):

    def __init__(self, name: str,
                 version: str,
                 start_date: datetime.date,
                 end_date: datetime.date,
                 trading_calendar: ExchangeCalendar,
                 frequency: datetime.timedelta | Period,
                 original_frequency: datetime.timedelta | Period,
                 data_type: DataType,
                 timestamp: datetime.datetime,
                 data: pl.DataFrame = None,
                 ):
        super().__init__(name=name,
                         start_date=start_date,
                         end_date=end_date,
                         frequency=frequency,
                         data_type=data_type,
                         original_frequency=original_frequency
                         )
        self.version = version
        self.start_date = start_date
        self.end_date = end_date
        self.trading_calendar = trading_calendar
        self.frequency = frequency
        self.frequency_td = period_to_timedelta(self.frequency)
        self.timestamp = timestamp
        self.data = data
        self._logger = structlog.get_logger(__name__)

    def get_dataframe(self) -> pl.DataFrame:
        return self.data

    @lru_cache
    def get_dataframe_with_columns(self, columns: frozenset[str]) -> pl.DataFrame:
        return self.data.select(pl.col(col) for col in columns)

    @lru_cache
    def get_dataframe_by_sid_and_columns(self, sid: str, columns: frozenset[str]) -> pl.DataFrame:
        return self.data.select(pl.col(col) for col in columns).filter(pl.col("sid").is_in(sid))

    def get_data_by_date(self, fields: frozenset[str],
                         from_date: datetime.datetime,
                         to_date: datetime.datetime,
                         frequency: datetime.timedelta | Period,
                         assets: frozenset[Asset],
                         include_bounds: bool,
                         ) -> pl.DataFrame:

        cols = set(fields.union({"date", "sid"}))
        if include_bounds:
            df = self.get_dataframe().select(pl.col(col) for col in cols).filter(
                pl.col("date") <= to_date,
                pl.col("date") >= from_date,
                pl.col("sid").is_in([asset.sid for asset in assets])
            ).group_by(pl.col("sid")).all()
        else:
            df = self.get_dataframe().select(pl.col(col) for col in cols).filter(
                pl.col("date") < to_date,
                pl.col("date") > from_date,
                pl.col("sid").is_in([asset.sid for asset in assets])).group_by(pl.col("sid")).all()
        if self.frequency < frequency:
            df = df.group_by_dynamic(
                index_column="date", every=frequency, by="sid").agg(pl.col(field).last() for field in fields)
        return df.sort(by="date")

    def get_missing_data_by_limit(self, fields: frozenset[str],
                                  limit: int,
                                  end_date: datetime.datetime,
                                  frequency: datetime.timedelta | Period,
                                  assets: frozenset[Asset],
                                  include_end_date: bool,
                                  ) -> pl.DataFrame:

        return self.missing_data_bundle_source.get_data_sync(
            symbols=[asset.get_symbol_by_exchange(None) for asset in assets], frequency=frequency,
            date_from=end_date - frequency * limit,
            date_to=end_date)

    # @lru_cache(maxsize=100)
    def get_data_by_limit(self, fields: frozenset[str] | None,
                          limit: int,
                          end_date: datetime.datetime,
                          frequency: datetime.timedelta | Period,
                          assets: frozenset[Asset],
                          include_end_date: bool,
                          ) -> pl.DataFrame:
        frequency_td = period_to_timedelta(frequency)
        total_bar_count = limit
        if end_date > self.end_date:
            raise ValueError(f"Requested end date {end_date} is greater than end date {self.end_date} of the bundle.")
            return self.get_missing_data_by_limit(frequency=frequency, assets=assets, fields=fields,
                                                  limit=limit, include_end_date=include_end_date,
                                                  end_date=end_date
                                                  )  # pl.DataFrame() # we have missing data

        if self.frequency_td < frequency_td:
            multiplier = int(frequency_td / self.frequency_td)
            total_bar_count = limit * multiplier
        df = self.get_dataframe()
        if fields is None:
            fields = frozenset(df.columns)
        cols = list(fields.union({"date", "sid"}))

        if include_end_date:
            df_raw = self.get_dataframe().select(pl.col(col) for col in cols).filter(
                pl.col("date") <= end_date,
                pl.col("sid").is_in([asset.sid for asset in assets])
            ).group_by(pl.col("sid")).tail(total_bar_count).sort(by="date")
        else:
            df_raw = self.get_dataframe().select(pl.col(col) for col in cols).filter(
                pl.col("date") < end_date,
                pl.col("sid").is_in([asset.sid for asset in assets])).group_by(pl.col("sid")).tail(
                total_bar_count).sort(by="date")

        if self.frequency_td < frequency_td:
            df = df_raw.group_by_dynamic(
                index_column="date", every=frequency, by="sid").agg(pl.col(field).last() for field in fields).tail(
                limit)
            return df
        return df_raw

    def get_spot_value(self, assets: frozenset[Asset], fields: frozenset[str], dt: datetime.datetime,
                       frequency: datetime.timedelta):
        """Public API method that returns a scalar value representing the value
        of the desired asset's field at either the given dt.

        Parameters
        ----------
        assets : Asset, ContinuousFuture, or iterable of same.
            The asset or assets whose data is desired.
        field : {'open', 'high', 'low', 'close', 'volume',
                 'price', 'last_traded'}
            The desired field of the asset.
        dt : datetime.datetime
            The timestamp for the desired value.
        data_frequency : str
            The frequency of the data to query; i.e. whether the data is
            'daily' or 'minute' bars

        Returns
        -------
        value : float, int, or datetime.datetime
            The spot value of ``field`` for ``asset`` The return type is based
            on the ``field`` requested. If the field is one of 'open', 'high',
            'low', 'close', or 'price', the value will be a float. If the
            ``field`` is 'volume' the value will be a int. If the ``field`` is
            'last_traded' the value will be a Timestamp.
        """
        # print(f"get spot value: {assets}, {fields}, {dt}")
        df_raw = self.get_data_by_limit(
            fields=fields,
            limit=1,
            end_date=dt,
            frequency=frequency,
            assets=assets,
            include_end_date=True,
        )
        return df_raw

    async def get_adjusted_value(
            self, asset: Asset, field: str, dt: datetime.datetime, perspective_dt: datetime.datetime,
            data_frequency: datetime.timedelta,
            spot_value: float = None
    ):
        """Returns a scalar value representing the value
        of the desired asset's field at the given dt with adjustments applied.

        Parameters
        ----------
        asset : Asset
            The asset whose data is desired.
        field : {'open', 'high', 'low', 'close', 'volume', \
                 'price', 'last_traded'}
            The desired field of the asset.
        dt : datetime.datetime
            The timestamp for the desired value.
        perspective_dt : datetime.datetime
            The timestamp from which the data is being viewed back from.
        data_frequency : str
            The frequency of the data to query; i.e. whether the data is
            'daily' or 'minute' bars

        Returns
        -------
        value : float, int, or datetime.datetime
            The value of the given ``field`` for ``asset`` at ``dt`` with any
            adjustments known by ``perspective_dt`` applied. The return type is
            based on the ``field`` requested. If the field is one of 'open',
            'high', 'low', 'close', or 'price', the value will be a float. If
            the ``field`` is 'volume' the value will be a int. If the ``field``
            is 'last_traded' the value will be a Timestamp.
        """
        if spot_value is None:
            spot_value = self.get_spot_value(assets=frozenset({asset}), fields=frozenset({field}), dt=dt,
                                             data_frequency=data_frequency)

        if isinstance(asset, Equity):  # TODO: fix this, not valid way to check if it is equity
            ratio = self.get_adjustments(assets=frozenset({asset}), field=field, dt=dt, perspective_dt=perspective_dt)[
                0]
            spot_value *= ratio

        return spot_value

    async def _get_adjustment_list(self, asset: Asset, adjustments_dict: dict[str, Any], table_name: str):
        """Internal method that returns a list of adjustments for the given sid.

        Parameters
        ----------
        asset : Asset
            The asset for which to return adjustments.

        adjustments_dict: dict
            A dictionary of sid -> list that is used as a cache.

        table_name: string
            The table that contains this data in the adjustments db.

        Returns
        -------
        adjustments: list
            A list of [multiplier, datetime.datetime], earliest first

        """
        if self.adjustment_repository is None:
            return []

        sid = asset.sid

        try:
            adjustments = adjustments_dict[sid]
        except KeyError:
            adjustments = adjustments_dict[
                sid
            ] = self.adjustment_repository.get_adjustments_for_sid(table_name, sid)

        return adjustments

    async def get_current_future_chain(self, continuous_future: ContinuousFuture, dt: datetime.datetime):
        """Retrieves the future chain for the contract at the given `dt` according
        the `continuous_future` specification.

        Returns
        -------

        future_chain : list[Future]
            A list of active futures, where the first index is the current
            contract specified by the continuous future definition, the second
            is the next upcoming contract and so on.
        """
        rf = self._roll_finders[continuous_future.roll_style]
        session = self.trading_calendar.minute_to_session(dt)
        contract_center = rf.get_contract_center(
            continuous_future.root_symbol, session, continuous_future.offset
        )
        oc = self.asset_repository.get_ordered_contracts(continuous_future.root_symbol)
        chain = oc.active_chain(contract_center, session.value)
        return self.asset_repository.retrieve_all(sids=chain)

    async def _get_current_contract(self, continuous_future: ContinuousFuture, dt: datetime.datetime):
        rf = self._roll_finders[continuous_future.roll_style]
        contract_sid = rf.get_contract_center(
            continuous_future.root_symbol, dt, continuous_future.offset
        )
        if contract_sid is None:
            return None
        return self.asset_repository.retrieve_asset(sid=contract_sid)

    async def get_adjustments(self, assets: frozenset[Asset], field: str, dt: datetime.datetime,
                              perspective_dt: datetime.datetime):
        """Returns a list of adjustments between the dt and perspective_dt for the
        given field and list of assets

        Parameters
        ----------
        assets : list of type Asset, or Asset
            The asset, or assets whose adjustments are desired.
        field : {'open', 'high', 'low', 'close', 'volume', \
                 'price', 'last_traded'}
            The desired field of the asset.
        dt : datetime.datetime
            The timestamp for the desired value.
        perspective_dt : datetime.datetime
            The timestamp from which the data is being viewed back from.

        Returns
        -------
        adjustments : list[Adjustment]
            The adjustments to that field.
        """
        adjustment_ratios_per_asset = []

        def split_adj_factor(x):
            return x if field != "volume" else 1.0 / x

        for asset in assets:
            adjustments_for_asset = []
            split_adjustments = self._get_adjustment_list(
                asset, self._splits_dict, "SPLITS"
            )
            for adj_dt, adj in split_adjustments:
                if dt < adj_dt.tz_localize(dt.tzinfo) <= perspective_dt:
                    adjustments_for_asset.append(split_adj_factor(adj))
                elif adj_dt.tz_localize(dt.tzinfo) > perspective_dt:
                    break

            if field != "volume":
                merger_adjustments = self._get_adjustment_list(
                    asset, self._mergers_dict, "MERGERS"
                )
                for adj_dt, adj in merger_adjustments:
                    if dt < adj_dt <= perspective_dt:
                        adjustments_for_asset.append(adj)
                    elif adj_dt > perspective_dt:
                        break

                dividend_adjustments = self._get_adjustment_list(
                    asset,
                    self._dividends_dict,
                    "DIVIDENDS",
                )
                for adj_dt, adj in dividend_adjustments:
                    if dt < adj_dt.tz_localize(dt.tzinfo) <= perspective_dt:
                        adjustments_for_asset.append(adj)
                    elif adj_dt.tz_localize(dt.tzinfo) > perspective_dt:
                        break

            ratio = reduce(mul, adjustments_for_asset, 1.0)
            adjustment_ratios_per_asset.append(ratio)

        return adjustment_ratios_per_asset



import datetime
import polars as pl
from abc import abstractmethod

from exchange_calendars import ExchangeCalendar

from ziplime.assets.entities.asset import Asset
from ziplime.data.domain.data_bundle import DataBundle
from ziplime.data.services.data_source import DataSource

from ziplime.domain.position import Position
from ziplime.domain.portfolio import Portfolio
from ziplime.domain.account import Account
from ziplime.finance.commission import CommissionModel
from ziplime.finance.domain.order import Order
# from ziplime.finance.slippage.slippage_model import SlippageModel
from ziplime.gens.domain.trading_clock import TradingClock


class Exchange(DataSource):

    def __init__(self, name: str, canonical_name: str, country_code: str,
                 clock: TradingClock,
                 trading_calendar: ExchangeCalendar,
                 data_bundle: DataBundle | None = None):
        self.name = name
        self.canonical_name = canonical_name
        self.country_code = country_code
        self.clock = clock
        self.trading_calendar = trading_calendar
        self.data_bundle = data_bundle

    def get_start_cash_balance(self):
        pass

    def get_current_cash_balance(self):
        pass

    def subscribe_to_market_data(self, asset):
        return []

    def get_subscribed_assets(self):
        return []

    @abstractmethod
    def get_positions(self) -> dict[Asset, Position]:
        ...

    @abstractmethod
    def get_portfolio(self) -> Portfolio:
        ...

    @abstractmethod
    def get_account(self) -> Account:
        ...

    @abstractmethod
    def get_time_skew(self):
        ...

    @abstractmethod
    async def submit_order(self, order: Order):
        ...

    def is_alive(self):
        ...

    @abstractmethod
    def get_orders(self) -> dict[str, Order]:
        ...

    @abstractmethod
    async def get_transactions(self, orders: dict[Asset, dict[str, Order]], current_dt: datetime.datetime):
        ...

    @abstractmethod
    def get_orders_by_ids(self, order_ids: list[str]):
        ...

    @abstractmethod
    def get_transactions_by_order_ids(self, order_ids: list[str]):
        ...

    @abstractmethod
    def cancel_order(self, order_id: str) -> None:
        ...

    @abstractmethod
    def get_last_traded_dt(self, asset):
        ...

    @abstractmethod
    def get_spot_value(self, assets, field, dt, data_frequency):
        ...

    @abstractmethod
    def get_realtime_bars(self, assets, frequency):
        ...

    @abstractmethod
    def get_slippage_model(self, asset: Asset):
        ...

    @abstractmethod
    def get_commission_model(self, asset: Asset) -> CommissionModel:
        ...

    @abstractmethod
    async def get_scalar_asset_spot_value(self, asset: Asset, field: str, dt: datetime.datetime,
                                          frequency: datetime.timedelta): ...

    @abstractmethod
    def get_scalar_asset_spot_value_sync(self, asset: Asset, field: str, dt: datetime.datetime,
                                         frequency: datetime.timedelta): ...

    async def get_spot_values(self, assets: frozenset[Asset], fields: frozenset[str], exchange_name: str): ...

    @abstractmethod
    def current(self, assets: frozenset[Asset], fields: frozenset[str], dt: datetime.datetime): ...

    def get_data_by_limit(self, fields: frozenset[str],
                          limit: int,
                          end_date: datetime.datetime,
                          frequency: datetime.timedelta,
                          assets: frozenset[Asset],
                          include_end_date: bool,
                          ) -> pl.DataFrame:
        ...



import datetime
import multiprocessing
import os
import sys
from typing import Self

import limexhub
import structlog
from asyncclick import progressbar
from joblib import Parallel, delayed

import polars as pl

from ziplime.data.data_sources.asset_data_source import AssetDataSource
from ziplime.data.services.data_bundle_source import DataBundleSource


class LimexHubAssetDataSource(AssetDataSource):
    def __init__(self, limex_api_key: str, maximum_threads: int | None = None):
        super().__init__()
        self._limex_api_key = limex_api_key
        self._logger = structlog.get_logger(__name__)
        self._limex_client = limexhub.RestAPI(token=limex_api_key)
        if maximum_threads is not None:
            self._maximum_threads = min(multiprocessing.cpu_count() * 2, maximum_threads)
        else:
            self._maximum_threads = multiprocessing.cpu_count() * 2

    async def get_assets(self, symbols: list[str], **kwargs) -> pl.DataFrame:
        assets = self._limex_client.instruments("SPX")
        return assets

    async def get_constituents(self, index: str) -> pl.DataFrame:
        assets = self._limex_client.constituents(index)
        return assets

    @classmethod
    def from_env(cls) -> Self:
        limex_hub_key = os.environ.get("LIMEX_API_KEY", None)
        maximum_threads = os.environ.get("LIMEX_HUB_MAXIMUM_THREADS", None)
        if limex_hub_key is None:
            raise ValueError("Missing LIMEX_API_KEY environment variable.")
        return cls(limex_api_key=limex_hub_key, maximum_threads=maximum_threads)



""" Utilities for working with numpy arrays."""

from collections import OrderedDict
from datetime import datetime

import numpy as np
from numpy.lib.stride_tricks import as_strided
from toolz import flip

bool_dtype = np.dtype("bool")
uint8_dtype = np.dtype("uint8")

uint32_dtype = np.dtype("uint32")
uint64_dtype = np.dtype("uint64")
int64_dtype = np.dtype("int64")

float32_dtype = np.dtype("float32")
float64_dtype = np.dtype("float64")

complex128_dtype = np.dtype("complex128")

datetime64D_dtype = np.dtype("datetime64[D]")
datetime64ns_dtype = np.dtype("datetime64[ns]")

object_dtype = np.dtype("O")
# We use object arrays for strings.
categorical_dtype = object_dtype

make_datetime64ns = flip(np.datetime64, "ns")
make_datetime64D = flip(np.datetime64, "D")

# Array compare that works across versions of numpy
try:
    assert_array_compare = np.testing.utils.assert_array_compare
except AttributeError:
    assert_array_compare = np.testing.assert_array_compare

NaTmap = {
    np.dtype(f"datetime64[{unit}]"): np.datetime64("NaT", unit)
    for unit in ("ns", "us", "ms", "s", "m", "D")
}


def NaT_for_dtype(dtype):
    """Retrieve NaT with the same units as ``dtype``.

    Parameters
    ----------
    dtype : dtype-coercable
        The dtype to lookup the NaT value for.

    Returns
    -------
    NaT : dtype
        The NaT value for the given dtype.
    """
    return NaTmap[np.dtype(dtype)]


NaTns = NaT_for_dtype(datetime64ns_dtype)
NaTD = NaT_for_dtype(datetime64D_dtype)

_FILLVALUE_DEFAULTS = {
    bool_dtype: False,
    float32_dtype: np.nan,
    float64_dtype: np.nan,
    datetime64ns_dtype: NaTns,
    object_dtype: None,
}

INT_DTYPES_BY_SIZE_BYTES = OrderedDict(
    [
        (1, np.dtype("int8")),
        (2, np.dtype("int16")),
        (4, np.dtype("int32")),
        (8, np.dtype("int64")),
    ]
)

UNSIGNED_INT_DTYPES_BY_SIZE_BYTES = OrderedDict(
    [
        (1, np.dtype("uint8")),
        (2, np.dtype("uint16")),
        (4, np.dtype("uint32")),
        (8, np.dtype("uint64")),
    ]
)


def int_dtype_with_size_in_bytes(size):
    try:
        return INT_DTYPES_BY_SIZE_BYTES[size]
    except KeyError as exc:
        raise ValueError("No integral dtype whose size is %d bytes." % size) from exc


def unsigned_int_dtype_with_size_in_bytes(size):
    try:
        return UNSIGNED_INT_DTYPES_BY_SIZE_BYTES[size]
    except KeyError as exc:
        raise ValueError(
            "No unsigned integral dtype whose size is %d bytes." % size
        ) from exc


class NoDefaultMissingValue(Exception):
    pass


def make_kind_check(python_types, numpy_kind):
    """
    Make a function that checks whether a scalar or array is of a given kind
    (e.g. float, int, datetime, timedelta).
    """

    def check(value):
        if hasattr(value, "dtype"):
            return value.dtype.kind == numpy_kind
        return isinstance(value, python_types)

    return check


is_float = make_kind_check(float, "f")
is_int = make_kind_check(int, "i")
is_datetime = make_kind_check(datetime, "M")
is_object = make_kind_check(object, "O")


def coerce_to_dtype(dtype, value):
    """
    Make a value with the specified numpy dtype.

    Only datetime64[ns] and datetime64[D] are supported for datetime dtypes.
    """
    name = dtype.name
    if name.startswith("datetime64"):
        if name == "datetime64[ns]":
            return make_datetime64ns(value)
        else:
            raise TypeError(f"Don't know how to coerce values of dtype {dtype}")
    return dtype.type(value)


def default_missing_value_for_dtype(dtype):
    """
    Get the default fill value for `dtype`.
    """
    try:
        return _FILLVALUE_DEFAULTS[dtype]
    except KeyError as exc:
        raise NoDefaultMissingValue(
            "No default value registered for dtype %s." % dtype
        ) from exc


def repeat_first_axis(array, count):
    """
    Restride `array` to repeat `count` times along the first axis.

    Parameters
    ----------
    array : np.array
        The array to restride.
    count : int
        Number of times to repeat `array`.

    Returns
    -------
    result : array
        Array of shape (count,) + array.shape, composed of `array` repeated
        `count` times along the first axis.

    Example
    -------
    >>> from numpy import arange
    >>> a = arange(3); a
    array([0, 1, 2])
    >>> repeat_first_axis(a, 2)
    array([[0, 1, 2],
           [0, 1, 2]])
    >>> repeat_first_axis(a, 4)
    array([[0, 1, 2],
           [0, 1, 2],
           [0, 1, 2],
           [0, 1, 2]])

    Notes
    ----
    The resulting array will share memory with `array`.  If you need to assign
    to the input or output, you should probably make a copy first.

    See Also
    --------
    repeat_last_axis
    """
    return as_strided(array, (count,) + array.shape, (0,) + array.strides)


def repeat_last_axis(array, count):
    """
    Restride `array` to repeat `count` times along the last axis.

    Parameters
    ----------
    array : np.array
        The array to restride.
    count : int
        Number of times to repeat `array`.

    Returns
    -------
    result : array
        Array of shape array.shape + (count,) composed of `array` repeated
        `count` times along the last axis.

    Example
    -------
    >>> from numpy import arange
    >>> a = arange(3); a
    array([0, 1, 2])
    >>> repeat_last_axis(a, 2)
    array([[0, 0],
           [1, 1],
           [2, 2]])
    >>> repeat_last_axis(a, 4)
    array([[0, 0, 0, 0],
           [1, 1, 1, 1],
           [2, 2, 2, 2]])

    Notes
    ----
    The resulting array will share memory with `array`.  If you need to assign
    to the input or output, you should probably make a copy first.

    See Also
    --------
    repeat_last_axis
    """
    return as_strided(array, array.shape + (count,), array.strides + (0,))


def rolling_window(array, length):
    """
    Restride an array of shape

        (X_0, ... X_N)

    into an array of shape

        (length, X_0 - length + 1, ... X_N)

    where each slice at index i along the first axis is equivalent to

        result[i] = array[length * i:length * (i + 1)]

    Parameters
    ----------
    array : np.ndarray
        The base array.
    length : int
        Length of the synthetic first axis to generate.

    Returns
    -------
    out : np.ndarray

    Example
    -------
    >>> from numpy import arange
    >>> a = arange(25).reshape(5, 5)
    >>> a
    array([[ 0,  1,  2,  3,  4],
           [ 5,  6,  7,  8,  9],
           [10, 11, 12, 13, 14],
           [15, 16, 17, 18, 19],
           [20, 21, 22, 23, 24]])

    >>> rolling_window(a, 2)
    array([[[ 0,  1,  2,  3,  4],
            [ 5,  6,  7,  8,  9]],
    <BLANKLINE>
           [[ 5,  6,  7,  8,  9],
            [10, 11, 12, 13, 14]],
    <BLANKLINE>
           [[10, 11, 12, 13, 14],
            [15, 16, 17, 18, 19]],
    <BLANKLINE>
           [[15, 16, 17, 18, 19],
            [20, 21, 22, 23, 24]]])
    """
    orig_shape = array.shape
    if not orig_shape:
        raise IndexError("Can't restride a scalar.")
    elif orig_shape[0] <= length:
        raise IndexError(
            "Can't restride array of shape {shape} with"
            " a window length of {len}".format(
                shape=orig_shape,
                len=length,
            )
        )

    num_windows = orig_shape[0] - length + 1
    new_shape = (num_windows, length) + orig_shape[1:]

    new_strides = (array.strides[0],) + array.strides

    return as_strided(array, new_shape, new_strides)


# Sentinel value that isn't NaT.
_notNaT = make_datetime64D(0)
iNaT = int(NaTns.view(int64_dtype))
assert iNaT == NaTD.view(int64_dtype), "iNaTns != iNaTD"


def isnat(obj):
    """
    Check if a value is np.NaT.
    """
    if obj.dtype.kind not in ("m", "M"):
        raise ValueError("%s is not a numpy datetime or timedelta")
    return obj.view(int64_dtype) == iNaT


def is_missing(data, missing_value):
    """
    Generic is_missing function that handles NaN and NaT.
    """
    if is_float(data) and np.isnan(missing_value):
        return np.isnan(data)
    elif is_datetime(data) and isnat(missing_value):
        return isnat(data)
    elif is_object(data) and missing_value is None:
        # XXX: Older versions of numpy returns True/False for array ==
        # None. Work around this by boxing None in a 1x1 array, which causes
        # numpy to do the broadcasted comparison we want.
        return data == np.array([missing_value])
    return data == missing_value


def same(x, y):
    """
    Check if two scalar values are "the same".

    Returns True if `x == y`, or if x and y are both NaN or both NaT.
    """
    if is_float(x) and np.isnan(x) and is_float(y) and np.isnan(y):
        return True
    elif is_datetime(x) and isnat(x) and is_datetime(y) and isnat(y):
        return True
    else:
        return x == y


def busday_count_mask_NaT(begindates, enddates, out=None):
    """
    Simple of numpy.busday_count that returns `float` arrays rather than int
    arrays, and handles `NaT`s by returning `NaN`s where the inputs were `NaT`.

    Doesn't support custom weekdays or calendars, but probably should in the
    future.

    See Also
    --------
    np.busday_count
    """
    if out is None:
        out = np.empty(np.broadcast(begindates, enddates).shape, dtype=float)

    beginmask = isnat(begindates)
    endmask = isnat(enddates)

    out = np.busday_count(
        # Temporarily fill in non-NaT values.
        np.where(beginmask, _notNaT, begindates),
        np.where(endmask, _notNaT, enddates),
        out=out,
    )

    # Fill in entries where either comparison was NaT with nan in the output.
    out[beginmask | endmask] = np.nan
    return out


def vectorized_is_element(array, choices):
    """
    Check if each element of ``array`` is in choices.

    Parameters
    ----------
    array : np.ndarray
    choices : object
        Object implementing __contains__.

    Returns
    -------
    was_element : np.ndarray[bool]
        Array indicating whether each element of ``array`` was in ``choices``.
    """
    return np.vectorize(choices.__contains__, otypes=[bool])(array)


def as_column(a):
    """
    Convert an array of shape (N,) into an array of shape (N, 1).

    This is equivalent to `a[:, np.newaxis]`.

    Parameters
    ----------
    a : np.ndarray

    Example
    -------
    >>> import numpy as np
    >>> a = np.arange(5)
    >>> a
    array([0, 1, 2, 3, 4])
    >>> as_column(a)
    array([[0],
           [1],
           [2],
           [3],
           [4]])
    >>> as_column(a).shape
    (5, 1)
    """
    if a.ndim != 1:
        raise ValueError(
            "as_column expected an 1-dimensional array, "
            "but got an array of shape %s" % (a.shape,)
        )
    return a[:, None]


def changed_locations(a, include_first):
    """
    Compute indices of values in ``a`` that differ from the previous value.

    Parameters
    ----------
    a : np.ndarray
        The array on which to indices of change.
    include_first : bool
        Whether or not to consider the first index of the array as "changed".

    Example
    -------
    >>> import numpy as np
    >>> changed_locations(np.array([0, 0, 5, 5, 1, 1]), include_first=False)
    array([2, 4])

    >>> changed_locations(np.array([0, 0, 5, 5, 1, 1]), include_first=True)
    array([0, 2, 4])
    """
    if a.ndim > 1:
        raise ValueError("indices_of_changed_values only supports 1D arrays.")
    indices = np.flatnonzero(np.diff(a)) + 1

    if not include_first:
        return indices

    return np.hstack([[0], indices])


def bytes_array_to_native_str_object_array(a):
    """Convert an array of dtype S to an object array containing `str`."""

    return a.astype(str).astype(object)



import datetime
import uuid
from uuid import UUID

from sqlalchemy import Numeric, String, TextClause, ForeignKey
from sqlalchemy.orm import mapped_column
from typing_extensions import Annotated

IntegerPK = Annotated[int, mapped_column(primary_key=True)]
UuidPK = Annotated[UUID, mapped_column(primary_key=True)]
StringPK = Annotated[str, mapped_column(primary_key=True)]
UuidUnique = Annotated[UUID, mapped_column(unique=True, default_factory=uuid.uuid4,
                                           server_default=TextClause("gen_random_uuid()"))]
StringUnique = Annotated[str, mapped_column(unique=True)]
String32 = Annotated[str, mapped_column(String(length=32))]
String64 = Annotated[str, mapped_column(String(length=64))]
String256 = Annotated[str, mapped_column(String(length=256))]
String2048 = Annotated[str, mapped_column(String(length=2048))]
String4096 = Annotated[str, mapped_column(String(length=4096))]
String2048Indexed = Annotated[str, mapped_column(String(length=2048), index=True)]
StringIndexed = Annotated[str, mapped_column(index=True)]
StringDefaultEmpty = Annotated[str, mapped_column(default="", index=True, server_default="")]
Decimal8 = Annotated[float, mapped_column(Numeric(precision=12, scale=8))]
Decimal4 = Annotated[float, mapped_column(Numeric(precision=8, scale=4))]
Decimal2 = Annotated[float, mapped_column(Numeric(precision=8, scale=2))]
DateTimeIndexed = Annotated[datetime.datetime, mapped_column(index=True)]
DateIndexed = Annotated[datetime.date, mapped_column(index=True)]

BooleanDefaultTrue = Annotated[bool, mapped_column(default=True, server_default="TRUE")]
BooleanDefaultFalse = Annotated[bool, mapped_column(default=False, server_default="FALSE")]
IntegerDefaultZero = Annotated[int, mapped_column(default=0, server_default="0")]
IntegerIndexed = Annotated[int, mapped_column(index=True)]

ExchangeFK = Annotated[str, mapped_column(ForeignKey("exchanges.exchange"), index=True)]
EquityFK = Annotated[int, mapped_column(ForeignKey("equities.sid"), index=True)]
AssetFK = Annotated[int, mapped_column(ForeignKey("assets.sid"), index=True)]
AssetRouterFK = Annotated[int, mapped_column(ForeignKey("asset_router.sid"), index=True)]
CurrencyFK = Annotated[int, mapped_column(ForeignKey("currencies.sid"), index=True)]

AssetRouterFKPK = Annotated[int, mapped_column(ForeignKey("asset_router.sid"), primary_key=True)]
SymbolsUniverseFKPK = Annotated[int, mapped_column(ForeignKey("symbols_universe.id"), primary_key=True)]

FuturesRootSymbolFK = Annotated[str, mapped_column(ForeignKey("futures_root_symbols.root_symbol"), index=True)]



# Copyright 2015 Quantopian, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from collections import defaultdict

from numpy import iinfo, uint32, multiply

from ziplime.data.fx import ExplodingFXRateReader
from ziplime.lib.adjusted_array import AdjustedArray
from ziplime.utils.numpy_utils import repeat_first_axis

from .base import PipelineLoader
from .utils import shift_dates
from ..data.equity_pricing import EquityPricing
from ...data.domain.data_bundle import DataBundle

UINT32_MAX = iinfo(uint32).max


class EquityPricingLoader(PipelineLoader):
    """A PipelineLoader for loading daily OHLCV data.

    Parameters
    ----------
    raw_price_reader : ziplime.data.session_bars.SessionBarReader
        Reader providing raw prices.
    adjustments_reader : ziplime.data.adjustments.SQLiteAdjustmentReader
        Reader providing price/volume adjustments.
    fx_reader : ziplime.data.fx.FXRateReader
       Reader providing currency conversions.
    """

    def __init__(self, raw_price_reader, adjustments_reader, fx_reader):
        self.raw_price_reader = raw_price_reader
        self.adjustments_reader = adjustments_reader
        self.fx_reader = fx_reader

    @classmethod
    def without_fx(cls, data_bundle: DataBundle):
        """
        Construct an EquityPricingLoader without support for fx rates.

        The returned loader will raise an error if requested to load
        currency-converted columns.

        Parameters
        ----------
        raw_price_reader : ziplime.data.session_bars.SessionBarReader
            Reader providing raw prices.
        adjustments_reader : ziplime.data.adjustments.SQLiteAdjustmentReader
            Reader providing price/volume adjustments.

        Returns
        -------
        loader : EquityPricingLoader
            A loader that can only provide currency-naive data.
        """
        return cls(
            raw_price_reader=None, # fix this
            adjustments_reader=None,
            fx_reader=ExplodingFXRateReader(),
        )

    def load_adjusted_array(self, domain, columns, dates, sids, mask):
        # load_adjusted_array is called with dates on which the user's algo
        # will be shown data, which means we need to return the data that would
        # be known at the **start** of each date. We assume that the latest
        # data known on day N is the data from day (N - 1), so we shift all
        # query dates back by a trading session.
        sessions = domain.sessions()
        shifted_dates = shift_dates(sessions, dates[0], dates[-1], shift=1)

        ohlcv_cols, currency_cols = self._split_column_types(columns)
        del columns  # From here on we should use ohlcv_cols or currency_cols.
        ohlcv_colnames = [c.name for c in ohlcv_cols]

        raw_ohlcv_arrays = self.raw_price_reader.load_raw_arrays(
            ohlcv_colnames,
            shifted_dates[0],
            shifted_dates[-1],
            sids,
        )

        # Currency convert raw_arrays in place if necessary. We use shifted
        # dates to load currency conversion rates to make them line up with
        # dates used to fetch prices.
        self._inplace_currency_convert(
            ohlcv_cols,
            raw_ohlcv_arrays,
            shifted_dates,
            sids,
        )

        adjustments = self.adjustments_reader.load_pricing_adjustments(
            ohlcv_colnames,
            dates,
            sids,
        )

        out = {}
        for c, c_raw, c_adjs in zip(ohlcv_cols, raw_ohlcv_arrays, adjustments):
            out[c] = AdjustedArray(
                c_raw.astype(c.dtype),
                c_adjs,
                c.missing_value,
            )

        for c in currency_cols:
            codes_1d = self.raw_price_reader.currency_codes(sids)
            codes = repeat_first_axis(codes_1d, len(dates))
            out[c] = AdjustedArray(
                codes,
                adjustments={},
                missing_value=None,
            )

        return out

    @property
    def currency_aware(self):
        # Tell the pipeline engine that this loader supports currency
        # conversion if we have a non-dummy fx rates reader.
        return not isinstance(self.fx_reader, ExplodingFXRateReader)

    def _inplace_currency_convert(self, columns, arrays, dates, sids):
        """
        Currency convert raw data loaded for ``column``.

        Parameters
        ----------
        columns : list[ziplime.pipeline.data.BoundColumn]
            List of columns whose raw data has been loaded.
        arrays : list[np.array]
            List of arrays, parallel to ``columns`` containing data for the
            column.
        dates : pd.DatetimeIndex
            Labels for rows of ``arrays``. These are the dates that should
            be used to fetch fx rates for conversion.
        sids : np.array[int64]
            Labels for columns of ``arrays``.

        Returns
        -------
        None

        Side Effects
        ------------
        Modifies ``arrays`` in place by applying currency conversions.
        """
        # Group columns by currency conversion spec.
        by_spec = defaultdict(list)
        for column, array in zip(columns, arrays):
            by_spec[column.currency_conversion].append(array)

        # Nothing to do for terms with no currency conversion.
        by_spec.pop(None, None)
        if not by_spec:
            return

        fx_reader = self.fx_reader
        base_currencies = self.raw_price_reader.currency_codes(sids)

        # Columns with the same conversion spec will use the same multipliers.
        for spec, arrays in by_spec.items():
            rates = fx_reader.get_rates(
                rate=spec.field,
                quote=spec.currency.code,
                bases=base_currencies,
                dts=dates,
            )
            for arr in arrays:
                multiply(arr, rates, out=arr)

    def _split_column_types(self, columns):
        """Split out currency columns from OHLCV columns.

        Parameters
        ----------
        columns : list[ziplime.pipeline.data.BoundColumn]
            Columns to be loaded by ``load_adjusted_array``.

        Returns
        -------
        ohlcv_columns : list[ziplime.pipeline.data.BoundColumn]
            Price and volume columns from ``columns``.
        currency_columns : list[ziplime.pipeline.data.BoundColumn]
            Currency code column from ``columns``, if present.
        """
        currency_name = EquityPricing.currency.name

        ohlcv = []
        currency = []
        for c in columns:
            if c.name == currency_name:
                currency.append(c)
            else:
                ohlcv.append(c)

        return ohlcv, currency


# Backwards compat alias.
USEquityPricingLoader = EquityPricingLoader



import datetime

from sqlalchemy.orm import Mapped

from ziplime.core.db.annotated_types import IntegerPK, StringIndexed, ExchangeFK, EquityFK, CurrencyFK
from ziplime.core.db.base_model import BaseModel


class CurrencySymbolMappingModel(BaseModel):
    __tablename__ = "currency_symbol_mappings"

    id: Mapped[IntegerPK]
    sid: Mapped[CurrencyFK]
    symbol: Mapped[str]
    start_date: Mapped[datetime.date]
    end_date: Mapped[datetime.date]
    exchange: Mapped[ExchangeFK | None]



from .base import FXRateReader, DEFAULT_FX_RATE
from .in_memory import InMemoryFXRateReader
from .exploding import ExplodingFXRateReader
from .hdf5 import HDF5FXRateReader, HDF5FXRateWriter

__all__ = [
    "DEFAULT_FX_RATE",
    "ExplodingFXRateReader",
    "FXRateReader",
    "HDF5FXRateReader",
    "HDF5FXRateWriter",
    "InMemoryFXRateReader",
]



"""
Factorization algorithms.
"""
import numpy as np

from ziplime.utils.numpy_utils import unsigned_int_dtype_with_size_in_bytes

def log2(d: float) -> float:
    return np.log(d) / np.log(2)

def smallest_uint_that_can_hold(maxval: int):
    """Choose the smallest numpy unsigned int dtype that can hold ``maxval``.
    """
    if maxval < 1:
        # lim x -> 0 log2(x) == -infinity so we floor at uint8
        return np.uint8
    else:
        # The number of bits required to hold the codes up to ``length`` is
        # log2(length). The number of bits per bytes is 8. We cannot have
        # fractional bytes so we need to round up. Finally, we can only have
        # integers with widths 1, 2, 4, or 8 so so we need to round up to the
        # next value by looking up the next largest size in ``_int_sizes``.
        return unsigned_int_dtype_with_size_in_bytes(
            _int_sizes[int(np.ceil(log2(maxval) / 8))]
        )

# ctypedef fused unsigned_integral:
#     np.uint8_t
#     np.uint16_t
#     np.uint32_t
#     np.uint64_t

class _NoneFirstSortKey:
    """Box to sort ``None`` to the front of the categories list.
    """

    def __init__(self, value):
        self.value = value

    def __eq__(self, other):
        return self.value == other.value

    def __lt__(self, other):
        if self.value is None:
            return True
        if other.value is None:
            return False
        return self.value < other.value

    def __gt__(self, other):
        if self.value is None:
            return False
        if other.value is None:
            return True
        return self.value > other.value

    def __le__(self, other):
        return self < other or self == other

    def __ge__(self, other):
        return self > other or self == other

def factorize_strings_known_impl(values: np.ndarray,
                                   nvalues: int,
                                  categories:list,
                                  missing_value,
                                  sort: int,
                                  codes:np.ndarray):
    if sort:
        categories = sorted(categories, key=lambda x: (x is None, x))

    reverse_categories = dict(        zip(categories, range(len(categories)))    )
    missing_code = reverse_categories[missing_value]

    for i in range(nvalues):
        codes[i] = reverse_categories.get(values[i], missing_code)

    return codes, np.asarray(categories, dtype=object), reverse_categories

def factorize_strings_known_categories(values: np.ndarray,
                                         categories: list,
                                         missing_value,
                                         sort:int):
    """
    Factorize an array whose categories are already known.

    Any entries not in the specified categories will be given the code for
    `missing_value`.
    """
    if missing_value not in categories:
        categories.insert(0, missing_value)

    ncategories = len(categories)
    nvalues = len(values)
    if ncategories <= 2 ** 8:
        return factorize_strings_known_impl(
            values,
            nvalues,
            categories,
            missing_value,
            sort,
            np.empty(nvalues, dtype=np.uint8)
        )
    elif ncategories <= 2 ** 16:
        return factorize_strings_known_impl(
            values,
            nvalues,
            categories,
            missing_value,
            sort,
            np.empty(nvalues, np.uint16),
        )
    elif ncategories <= 2 ** 32:
        return factorize_strings_known_impl(
            values,
            nvalues,
            categories,
            missing_value,
            sort,
            np.empty(nvalues, np.uint32),
        )
    elif ncategories <= 2 ** 64:
        return factorize_strings_known_impl(
            values,
            nvalues,
            categories,
            missing_value,
            sort,
            np.empty(nvalues, np.uint64),
        )
    else:
        raise ValueError('ncategories larger than uint64')

def factorize_strings_impl( values: np.ndarray,
                            missing_value,
                            sort: int,
                            codes: np.ndarray):
    categories = [missing_value]
    reverse_categories = {missing_value: 0}

    key = None

    for i in range(len(values)):
        key = values[i]
        code = reverse_categories.get(key, -1)
        if code == -1:
            # Assign new code.
            code = len(reverse_categories)
            reverse_categories[key] = code
            categories.append(key)
        codes[i] = code

     # np.ndarray[np.int64_t, ndim=1] sorter
    # cdef np.ndarray[unsigned_integral, ndim=1] reverse_indexer
    # cdef int ncategories
    categories_array = np.asarray(
        categories,
        dtype=object,
    )

    if sort:
        # This is all adapted from pandas.core.algorithms.factorize.
        ncategories = len(categories_array)
        sorter = np.empty(ncategories, dtype=np.int64)

        # Don't include missing_value in the argsort, because None is
        # unorderable with bytes/str in py3. Always just sort it to 0.
        sorter[1:] = categories_array[1:].argsort() + 1
        sorter[0] = 0

        reverse_indexer = np.empty(ncategories, dtype=codes.dtype)
        reverse_indexer.put(sorter, np.arange(ncategories))

        codes = reverse_indexer.take(codes)
        categories_array = categories_array.take(sorter)
        reverse_categories = dict(zip(categories_array, range(ncategories)))

    return codes, categories_array, reverse_categories

_int_sizes = [1, 1, 2, 4, 4, 8, 8, 8, 8]

def factorize_strings(values: np.ndarray,
                        missing_value,
                        sort: int):
    """
    Factorize an array of (possibly duplicated) labels into an array of indices
    into a unique array of labels.

    This is ~30% faster than pandas.factorize, at the cost of not having
    special treatment for NaN, which we don't care about because we only
    support arrays of strings.

    (Though it's faster even if you throw in the nan checks that pandas does,
    because we're using dict and list instead of PyObjectHashTable and
    ObjectVector.  Python's builtin data structures are **really**
    well-optimized.)
    """
    nvalues = len(values)

    # use exclusive less than because we need to account for the possibility
    # that the missing value is not in values
    if nvalues < 2 ** 8:
        # we won't try to shrink because the ``codes`` array cannot get any
        # smaller
        return factorize_strings_impl(
            values,
            missing_value,
            sort,
            np.empty(nvalues, dtype=np.uint8)
        )
    elif nvalues < 2 ** 16:
        (codes,
         categories_array,
         reverse_categories) = factorize_strings_impl(
            values,
            missing_value,
            sort,
            np.empty(nvalues, dtype=np.uint16),
        )
    elif nvalues < 2 ** 32:
        (codes,
         categories_array,
         reverse_categories) = factorize_strings_impl(
            values,
            missing_value,
            sort,
            np.empty(nvalues, dtype=np.uint32),
        )
    elif nvalues < 2 ** 64:
        (codes,
         categories_array,
         reverse_categories) = factorize_strings_impl(
            values,
            missing_value,
            sort,
            np.empty(nvalues, dtype=np.uint64),
        )
    else:
        # unreachable
        raise ValueError('nvalues larger than uint64')

    length = len(categories_array)
    narrowest_dtype = smallest_uint_that_can_hold(length)
    if codes.dtype != narrowest_dtype:
        # condense the codes down to the narrowest dtype possible
        codes = codes.astype(narrowest_dtype)

    return codes, categories_array, reverse_categories



"""
Dependency-Graph representation of Pipeline API terms.
"""

import uuid

import networkx as nx
from ziplime.pipeline.visualize import display_graph

from .term import LoadableTerm


class CyclicDependency(Exception):
    pass


# This sentinel value is uniquely-generated at import time so that we can
# guarantee that it never conflicts with a user-provided column name.
#
# (Yes, technically, a user can import this file and pass this as the name of a
# column. If you do that you deserve whatever bizarre failure you cause.)
SCREEN_NAME = "screen_" + uuid.uuid4().hex


class TermGraph:
    """
    An abstract representation of Pipeline Term dependencies.

    This class does not keep any additional metadata about any term relations
    other than dependency ordering.  As such it is only useful in contexts
    where you care exclusively about order properties (for example, when
    drawing visualizations of execution order).

    Parameters
    ----------
    terms : dict
        A dict mapping names to final output terms.

    Attributes
    ----------
    outputs

    Methods
    -------
    ordered()
        Return a topologically-sorted iterator over the terms in self.
    execution_order(workspace, refcounts)
        Return a topologically-sorted iterator over the terms in self, skipping
        entries in ``workspace`` and entries with refcounts of zero.

    See Also
    --------
    ExecutionPlan
    """

    def __init__(self, terms):
        self.graph = nx.DiGraph()

        self._frozen = False
        parents = set()
        for term in terms.values():
            self._add_to_graph(term, parents)
            # No parents should be left between top-level terms.
            assert not parents

        self._outputs = terms

        # Mark that no more terms should be added to the graph.
        self._frozen = True

    def __contains__(self, term):
        return term in self.graph

    def _add_to_graph(self, term, parents):
        """
        Add a term and all its children to ``graph``.

        ``parents`` is the set of all the parents of ``term` that we've added
        so far. It is only used to detect dependency cycles.
        """
        if self._frozen:
            raise ValueError(
                "Can't mutate %s after construction." % type(self).__name__
            )

        # If we've seen this node already as a parent of the current traversal,
        # it means we have an unsatisifiable dependency.  This should only be
        # possible if the term's inputs are mutated after construction.
        if term in parents:
            raise CyclicDependency(term)

        parents.add(term)

        self.graph.add_node(term)

        for dependency in term.dependencies:
            self._add_to_graph(dependency, parents)
            self.graph.add_edge(dependency, term)

        parents.remove(term)

    @property
    def outputs(self):
        """
        Dict mapping names to designated output terms.
        """
        return self._outputs

    @property
    def screen_name(self):
        """Name of the specially-designated ``screen`` term for the pipeline."""
        return SCREEN_NAME

    def execution_order(self, workspace, refcounts):
        """
        Return a topologically-sorted list of the terms in ``self`` which
        need to be computed.

        Filters out any terms that are already present in ``workspace``, as
        well as any terms with refcounts of 0.

        Parameters
        ----------
        workspace : dict[Term, np.ndarray]
            Initial state of workspace for a pipeline execution. May contain
            pre-computed values provided by ``populate_initial_workspace``.
        refcounts : dict[Term, int]
            Reference counts for terms to be computed. Terms with reference
            counts of 0 do not need to be computed.
        """
        return list(
            nx.topological_sort(
                self.graph.subgraph(
                    {
                        term
                        for term, refcount in refcounts.items()
                        if refcount > 0 and term not in workspace
                    },
                ),
            )
        )

    def ordered(self):
        return iter(nx.topological_sort(self.graph))

    #@lazyval
    @property
    def loadable_terms(self):
        return {term for term in self.graph if isinstance(term, LoadableTerm)}

    #@lazyval
    @property
    def jpeg(self):
        return display_graph(self, "jpeg")

    #@lazyval
    @property
    def png(self):
        return display_graph(self, "png")

    #@lazyval
    @property
    def svg(self):
        return display_graph(self, "svg")

    def _repr_png_(self):
        return self.png.data

    def initial_refcounts(self, initial_terms):
        """
        Calculate initial refcounts for execution of this graph.

        Parameters
        ----------
        initial_terms : iterable[Term]
            An iterable of terms that were pre-computed before graph execution.

        Each node starts with a refcount equal to its outdegree, and output
        nodes get one extra reference to ensure that they're still in the graph
        at the end of execution.
        """
        refcounts = dict(self.graph.out_degree())
        for t in self.outputs.values():
            refcounts[t] += 1

        for t in initial_terms:
            self._decref_dependencies_recursive(t, refcounts, set())

        return refcounts

    def _decref_dependencies_recursive(self, term, refcounts, garbage):
        """
        Decrement terms recursively.

        Notes
        -----
        This should only be used to build the initial workspace, after that we
        should use:
        :meth:`~ziplime.pipeline.graph.TermGraph.decref_dependencies`
        """
        # Edges are tuple of (from, to).
        for parent, _ in self.graph.in_edges([term]):
            refcounts[parent] -= 1
            # No one else depends on this term. Remove it from the
            # workspace to conserve memory.
            if refcounts[parent] == 0:
                garbage.add(parent)
                self._decref_dependencies_recursive(parent, refcounts, garbage)

    def decref_dependencies(self, term, refcounts):
        """
        Decrement in-edges for ``term`` after computation.

        Parameters
        ----------
        term : ziplime.pipeline.Term
            The term whose parents should be decref'ed.
        refcounts : dict[Term -> int]
            Dictionary of refcounts.

        Return
        ------
        garbage : set[Term]
            Terms whose refcounts hit zero after decrefing.
        """
        garbage = set()
        # Edges are tuple of (from, to).
        for parent, _ in self.graph.in_edges([term]):
            refcounts[parent] -= 1
            # No one else depends on this term. Remove it from the
            # workspace to conserve memory.
            if refcounts[parent] == 0:
                garbage.add(parent)
        return garbage

    def __len__(self):
        return len(self.graph)


class ExecutionPlan(TermGraph):
    """
    Graph representation of Pipeline Term dependencies that includes metadata
    about extra rows required to perform computations.

    Each node in the graph has an `extra_rows` attribute, indicating how many,
    if any, extra rows we should compute for the node. Extra rows are most
    often needed when a term is an input to a rolling window computation. For
    example, if we compute a 30 day moving average of price from day X to day
    Y, we need to load price data for the range from day (X - 29) to day Y.

    Parameters
    ----------
    domain : ziplime.pipeline.domain.Domain
        The domain of execution for which we need to build a plan.
    terms : dict
        A dict mapping names to final output terms.
    start_date : pd.Timestamp
        The first date for which output is requested for ``terms``.
    end_date : pd.Timestamp
        The last date for which output is requested for ``terms``.

    Attributes
    ----------
    domain
    extra_rows
    outputs
    offset
    """

    def __init__(self, domain, terms, start_date, end_date, min_extra_rows=0):
        super(ExecutionPlan, self).__init__(terms)

        # Specialize all the LoadableTerms in the graph to our domain, so that
        # when the engine requests an execution order, we emit the specialized
        # versions of loadable terms.
        #
        # NOTE: We're explicitly avoiding using self.loadable_terms here.
        #
        # At this point the graph still contains un-specialized loadable terms,
        # and this is where we're actually going through and specializing all
        # of them. We don't want use self.loadable_terms because it's a
        # lazyval, and we don't want its result to be cached until after we've
        # specialized.
        specializations = {
            t: t.specialize(domain) for t in self.graph if isinstance(t, LoadableTerm)
        }
        self.graph = nx.relabel.relabel_nodes(self.graph, specializations)

        self.domain = domain

        sessions = domain.sessions()
        for term in terms.values():
            self.set_extra_rows(
                term,
                sessions,
                start_date,
                end_date,
                min_extra_rows=min_extra_rows,
            )

        self._assert_all_loadable_terms_specialized_to(domain)

    def set_extra_rows(self, term, all_dates, start_date, end_date, min_extra_rows):
        # Specialize any loadable terms before adding extra rows.
        term = maybe_specialize(term, self.domain)

        # A term can require that additional extra rows beyond the minimum be
        # computed.  This is most often used with downsampled terms, which need
        # to ensure that the first date is a computation date.
        extra_rows_for_term = term.compute_extra_rows(
            all_dates,
            start_date,
            end_date,
            min_extra_rows,
        )
        if extra_rows_for_term < min_extra_rows:
            raise ValueError(
                "term %s requested fewer rows than the minimum of %d"
                % (
                    term,
                    min_extra_rows,
                )
            )

        self._ensure_extra_rows(term, extra_rows_for_term)

        for dependency, additional_extra_rows in term.dependencies.items():
            self.set_extra_rows(
                dependency,
                all_dates,
                start_date,
                end_date,
                min_extra_rows=extra_rows_for_term + additional_extra_rows,
            )

    #@lazyval
    @property
    def offset(self):
        """
        For all pairs (term, input) such that `input` is an input to `term`,
        compute a mapping::

            (term, input) -> offset(term, input)

        where ``offset(term, input)`` is the number of rows that ``term``
        should truncate off the raw array produced for ``input`` before using
        it. We compute this value as follows::

            offset(term, input) = (extra_rows_computed(input)
                                   - extra_rows_computed(term)
                                   - requested_extra_rows(term, input))
        Examples
        --------

        Case 1
        ~~~~~~

        Factor A needs 5 extra rows of USEquityPricing.close, and Factor B
        needs 3 extra rows of the same.  Factor A also requires 5 extra rows of
        USEquityPricing.high, which no other Factor uses.  We don't require any
        extra rows of Factor A or Factor B

        We load 5 extra rows of both `price` and `high` to ensure we can
        service Factor A, and the following offsets get computed::

            offset[Factor A, USEquityPricing.close] == (5 - 0) - 5 == 0
            offset[Factor A, USEquityPricing.high]  == (5 - 0) - 5 == 0
            offset[Factor B, USEquityPricing.close] == (5 - 0) - 3 == 2
            offset[Factor B, USEquityPricing.high] raises KeyError.

        Case 2
        ~~~~~~

        Factor A needs 5 extra rows of USEquityPricing.close, and Factor B
        needs 3 extra rows of Factor A, and Factor B needs 2 extra rows of
        USEquityPricing.close.

        We load 8 extra rows of USEquityPricing.close (enough to load 5 extra
        rows of Factor A), and the following offsets get computed::

            offset[Factor A, USEquityPricing.close] == (8 - 3) - 5 == 0
            offset[Factor B, USEquityPricing.close] == (8 - 0) - 2 == 6
            offset[Factor B, Factor A]              == (3 - 0) - 3 == 0

        Notes
        -----
        `offset(term, input) >= 0` for all valid pairs, since `input` must be
        an input to `term` if the pair appears in the mapping.

        This value is useful because we load enough rows of each input to serve
        all possible dependencies.  However, for any given dependency, we only
        want to compute using the actual number of required extra rows for that
        dependency.  We can do so by truncating off the first `offset` rows of
        the loaded data for `input`.

        See Also
        --------
        :meth:`ziplime.pipeline.graph.ExecutionPlan.offset`
        :meth:`ziplime.pipeline.engine.ExecutionPlan.mask_and_dates_for_term`
        :meth:`ziplime.pipeline.engine.SimplePipelineEngine._inputs_for_term`
        """
        extra = self.extra_rows

        out = {}
        for term in self.graph:
            for dep, requested_extra_rows in term.dependencies.items():
                specialized_dep = maybe_specialize(dep, self.domain)

                # How much bigger is the result for dep compared to term?
                size_difference = extra[specialized_dep] - extra[term]

                # Subtract the portion of that difference that was required by
                # term's lookback window.
                offset = size_difference - requested_extra_rows
                out[term, specialized_dep] = offset

        return out

    #@lazyval
    @property
    def extra_rows(self):
        """
        A dict mapping `term` -> `# of extra rows to load/compute of `term`.

        Notes
        ----
        This value depends on the other terms in the graph that require `term`
        **as an input**.  This is not to be confused with `term.dependencies`,
        which describes how many additional rows of `term`'s inputs we need to
        load, and which is determined entirely by `Term` itself.

        Examples
        --------
        Our graph contains the following terms:

            A = SimpleMovingAverage([USEquityPricing.high], window_length=5)
            B = SimpleMovingAverage([USEquityPricing.high], window_length=10)
            C = SimpleMovingAverage([USEquityPricing.low], window_length=8)

        To compute N rows of A, we need N + 4 extra rows of `high`.
        To compute N rows of B, we need N + 9 extra rows of `high`.
        To compute N rows of C, we need N + 7 extra rows of `low`.

        We store the following extra_row requirements:

        self.extra_rows[high] = 9  # Ensures that we can service B.
        self.extra_rows[low] = 7

        See Also
        --------
        :meth:`ziplime.pipeline.graph.ExecutionPlan.offset`
        :meth:`ziplime.pipeline.Term.dependencies`
        """

        return {term: self.graph.nodes[term]["extra_rows"] for term in self.graph.nodes}

    def _ensure_extra_rows(self, term, N):
        """
        Ensure that we're going to compute at least N extra rows of `term`.
        """
        attrs = dict(self.graph.nodes())[term]
        attrs["extra_rows"] = max(N, attrs.get("extra_rows", 0))

    def mask_and_dates_for_term(self, term, root_mask_term, workspace, all_dates):
        """
        Load mask and mask row labels for term.

        Parameters
        ----------
        term : Term
            The term to load the mask and labels for.
        root_mask_term : Term
            The term that represents the root asset exists mask.
        workspace : dict[Term, any]
            The values that have been computed for each term.
        all_dates : pd.DatetimeIndex
            All of the dates that are being computed for in the pipeline.

        Returns
        -------
        mask : np.ndarray
            The correct mask for this term.
        dates : np.ndarray
            The slice of dates for this term.
        """
        mask = term.mask
        mask_offset = self.extra_rows[mask] - self.extra_rows[term]

        # This offset is computed against root_mask_term because that is what
        # determines the shape of the top-level dates array.
        dates_offset = self.extra_rows[root_mask_term] - self.extra_rows[term]

        return workspace[mask][mask_offset:], all_dates[dates_offset:]

    def _assert_all_loadable_terms_specialized_to(self, domain):
        """Make sure that we've specialized all loadable terms in the graph."""
        for term in self.graph.nodes():
            if isinstance(term, LoadableTerm):
                assert term.domain is domain


# XXX: This function exists because we currently only specialize LoadableTerms
#      when running a Pipeline on a given domain.
def maybe_specialize(term, domain):
    """Specialize a term if it's loadable."""
    if isinstance(term, LoadableTerm):
        return term.specialize(domain)
    return term



import dataclasses
import datetime
from textwrap import dedent


class ZiplineError(Exception):
    msg = None

    def __init__(self, **kwargs):
        self.kwargs = kwargs

    @property
    def message(self):
        return str(self)

    def __str__(self):
        msg = self.msg.format(**self.kwargs)
        return msg

    __repr__ = __str__


class NoTradeDataAvailable(ZiplineError):
    pass


class NoTradeDataAvailableTooEarly(NoTradeDataAvailable):
    msg = "{sid} does not exist on {dt}. It started trading on {start_dt}."


class NoTradeDataAvailableTooLate(NoTradeDataAvailable):
    msg = "{sid} does not exist on {dt}. It stopped trading on {end_dt}."


class BenchmarkAssetNotAvailableTooEarly(NoTradeDataAvailableTooEarly):
    pass


class BenchmarkAssetNotAvailableTooLate(NoTradeDataAvailableTooLate):
    pass


class InvalidBenchmarkAsset(ZiplineError):
    msg = """
{sid} cannot be used as the benchmark because it has a stock \
dividend on {dt}.  Choose another asset to use as the benchmark.
""".strip()


class WrongDataForTransform(ZiplineError):
    """Raised whenever a rolling transform is called on an event that
    does not have the necessary properties.
    """

    msg = "{transform} requires {fields}. Event cannot be processed."


class UnsupportedSlippageModel(ZiplineError):
    """Raised if a user script calls the set_slippage magic
    with a slipage object that isn't a VolumeShareSlippage or
    FixedSlipapge
    """

    msg = """
You attempted to set slippage with an unsupported class. \
Please use VolumeShareSlippage or FixedSlippage.
""".strip()


class IncompatibleSlippageModel(ZiplineError):
    """Raised if a user tries to set a futures slippage model for equities or vice
    versa.
    """

    msg = """
You attempted to set an incompatible slippage model for {asset_type}. \
The slippage model '{given_model}' only supports {supported_asset_types}.
""".strip()


class SetSlippagePostInit(ZiplineError):
    # Raised if a users script calls set_slippage magic
    # after the initialize method has returned.
    msg = """
You attempted to set slippage outside of `initialize`. \
You may only call 'set_slippage' in your initialize method.
""".strip()


class SetCancelPolicyPostInit(ZiplineError):
    # Raised if a users script calls set_cancel_policy
    # after the initialize method has returned.
    msg = """
You attempted to set the cancel policy outside of `initialize`. \
You may only call 'set_cancel_policy' in your initialize method.
""".strip()


class RegisterTradingControlPostInit(ZiplineError):
    # Raised if a user's script register's a trading control after initialize
    # has been run.
    msg = """
You attempted to set a trading control outside of `initialize`. \
Trading controls may only be set in your initialize method.
""".strip()


class RegisterAccountControlPostInit(ZiplineError):
    # Raised if a user's script register's a trading control after initialize
    # has been run.
    msg = """
You attempted to set an account control outside of `initialize`. \
Account controls may only be set in your initialize method.
""".strip()


class UnsupportedCommissionModel(ZiplineError):
    """Raised if a user script calls the set_commission magic
    with a commission object that isn't a PerShare, PerTrade or
    PerDollar commission
    """

    msg = """
You attempted to set commission with an unsupported class. \
Please use PerShare or PerTrade.
""".strip()


class IncompatibleCommissionModel(ZiplineError):
    """Raised if a user tries to set a futures commission model for equities or
    vice versa.
    """

    msg = """
You attempted to set an incompatible commission model for {asset_type}. \
The commission model '{given_model}' only supports {supported_asset_types}.
""".strip()


class UnsupportedCancelPolicy(ZiplineError):
    """Raised if a user script calls set_cancel_policy with an object that isn't
    a CancelPolicy.
    """

    msg = """
You attempted to set the cancel policy with an unsupported class.  Please use
an instance of CancelPolicy.
""".strip()


class SetCommissionPostInit(ZiplineError):
    """Raised if a users script calls set_commission magic
    after the initialize method has returned.
    """

    msg = """
You attempted to override commission outside of `initialize`. \
You may only call 'set_commission' in your initialize method.
""".strip()


class TransactionWithNoVolume(ZiplineError):
    """Raised if a transact call returns a transaction with zero volume."""

    msg = """
Transaction {txn} has a volume of zero.
""".strip()


class TransactionWithWrongDirection(ZiplineError):
    """Raised if a transact call returns a transaction with a direction that
    does not match the order.
    """

    msg = """
Transaction {txn} not in same direction as corresponding order {order}.
""".strip()


class TransactionWithNoAmount(ZiplineError):
    """Raised if a transact call returns a transaction with zero amount."""

    msg = """
Transaction {txn} has an amount of zero.
""".strip()


class TransactionVolumeExceedsOrder(ZiplineError):
    """Raised if a transact call returns a transaction with a volume greater than
    the corresponding order.
    """

    msg = """
Transaction volume of {txn} exceeds the order volume of {order}.
""".strip()


class UnsupportedOrderParameters(ZiplineError):
    """Raised if a set of mutually exclusive parameters are passed to an order
    call.
    """

    msg = "{msg}"


class CannotOrderDelistedAsset(ZiplineError):
    """Raised if an order is for a delisted asset."""

    msg = "{msg}"


class BadOrderParameters(ZiplineError):
    """Raised if any impossible parameters (nan, negative limit/stop)
    are passed to an order call.
    """

    msg = "{msg}"


class OrderDuringInitialize(ZiplineError):
    """Raised if order is called during initialize()"""

    msg = "{msg}"


class SetBenchmarkOutsideInitialize(ZiplineError):
    """Raised if set_benchmark is called outside initialize()"""

    msg = "'set_benchmark' can only be called within initialize function."


class ZeroCapitalError(ZiplineError):
    """Raised if initial capital is set at or below zero"""

    msg = "initial capital base must be greater than zero"


class AccountControlViolation(ZiplineError):
    """Raised if the account violates a constraint set by a AccountControl."""

    msg = """
Account violates account constraint {constraint}.
""".strip()


class TradingControlViolation(ZiplineError):
    """Raised if an order would violate a constraint set by a TradingControl."""

    msg = """
Order for {amount} shares of {asset} at {datetime} violates trading constraint
{constraint}.
""".strip()


class IncompatibleHistoryFrequency(ZiplineError):
    """Raised when a frequency is given to history which is not supported.
    At least, not yet.
    """

    msg = """
Requested history at frequency '{frequency}' cannot be created with data
at frequency '{data_frequency}'.
""".strip()


class OrderInBeforeTradingStart(ZiplineError):
    """Raised when an algorithm calls an order method in before_trading_start."""

    msg = "Cannot place orders inside before_trading_start."


class MultipleSymbolsFound(ZiplineError):
    """Raised when a symbol() call contains a symbol that changed over
    time and is thus not resolvable without additional information
    provided via as_of_date.
    """

    msg = """
Multiple symbols with the name '{symbol}' found. Use the
as_of_date' argument to specify when the date symbol-lookup
should be valid.

Possible options: {options}
    """.strip()


class MultipleSymbolsFoundForFuzzySymbol(MultipleSymbolsFound):
    """Raised when a fuzzy symbol lookup is not resolvable without additional
    information.
    """

    msg = dedent(
        """\
        Multiple symbols were found fuzzy matching the name '{symbol}'. Use
        the as_of_date and/or country_code arguments to specify the date
        and country for the symbol-lookup.

        Possible options: {options}
    """
    )


class SameSymbolUsedAcrossCountries(MultipleSymbolsFound):
    """Raised when a symbol() call contains a symbol that is used in more than
    one country and is thus not resolvable without a country_code.
    """

    msg = dedent(
        """\
        The symbol '{symbol}' is used in more than one country. Use the
        country_code argument to specify the country.

        Possible options by country: {options}
    """
    )


class SymbolNotFound(ZiplineError):
    """Raised when a symbol() call contains a non-existant symbol."""

    msg = """
Symbol '{symbol}' was not found.
""".strip()


class RootSymbolNotFound(ZiplineError):
    """Raised when a lookup_future_chain() call contains a non-existant symbol."""

    msg = """
Root symbol '{root_symbol}' was not found.
""".strip()


class ValueNotFoundForField(ZiplineError):
    """Raised when a lookup_by_supplementary_mapping() call contains a
    value does not exist for the specified mapping type.
    """

    msg = """
Value '{value}' was not found for field '{field}'.
""".strip()


class MultipleValuesFoundForField(ZiplineError):
    """Raised when a lookup_by_supplementary_mapping() call contains a
    value that changed over time for the specified field and is
    thus not resolvable without additional information provided via
    as_of_date.
    """

    msg = """
Multiple occurrences of the value '{value}' found for field '{field}'.
Use the 'as_of_date' or 'country_code' argument to specify when or where the
lookup should be valid.

Possible options: {options}
    """.strip()


class NoValueForSid(ZiplineError):
    """Raised when a get_supplementary_field() call contains a sid that
    does not have a value for the specified mapping type.
    """

    msg = """
No '{field}' value found for sid '{sid}'.
""".strip()


class MultipleValuesFoundForSid(ZiplineError):
    """Raised when a get_supplementary_field() call contains a value that
    changed over time for the specified field and is thus not resolvable
    without additional information provided via as_of_date.
    """

    msg = """
Multiple '{field}' values found for sid '{sid}'. Use the as_of_date' argument
to specify when the lookup should be valid.

Possible options: {options}
""".strip()


class SidsNotFound(ZiplineError):
    """Raised when a retrieve_asset() or retrieve_all() call contains a
    non-existent sid.
    """

    @property
    def plural(self):
        return len(self.sids) > 1

    @property
    def sids(self):
        return self.kwargs["sids"]

    @property
    def msg(self):
        if self.plural:
            return "No assets found for sids: {sids}."
        return "No asset found for sid: {sids[0]}."


class EquitiesNotFound(SidsNotFound):
    """Raised when a call to `retrieve_equities` fails to find an asset."""

    @property
    def msg(self):
        if self.plural:
            return "No equities found for sids: {sids}."
        return "No equity found for sid: {sids[0]}."


class FutureContractsNotFound(SidsNotFound):
    """Raised when a call to `retrieve_futures_contracts` fails to find an asset."""

    @property
    def msg(self):
        if self.plural:
            return "No future contracts found for sids: {sids}."
        return "No future contract found for sid: {sids[0]}."


class ConsumeAssetMetaDataError(ZiplineError):
    """Raised when AssetFinder.consume() is called on an invalid object."""

    msg = """
AssetFinder can not consume metadata of type {obj}. Metadata must be a dict, a
DataFrame, or a tables.Table. If the provided metadata is a Table, the rows
must contain both or one of 'sid' or 'symbol'.
""".strip()


class SidAssignmentError(ZiplineError):
    """Raised when an AssetFinder tries to build an Asset that does not have a sid
    and that AssetFinder is not permitted to assign sids.
    """

    msg = """
AssetFinder metadata is missing a SID for identifier '{identifier}'.
""".strip()


class NoSourceError(ZiplineError):
    """Raised when no source is given to the pipeline"""

    msg = """
No data source given.
""".strip()


class PipelineDateError(ZiplineError):
    """Raised when only one date is passed to the pipeline"""

    msg = """
Only one simulation date given. Please specify both the 'start' and 'end' for
the simulation, or neither. If neither is given, the start and end of the
DataSource will be used. Given start = '{start}', end = '{end}'
""".strip()


class WindowLengthTooLong(ZiplineError):
    """Raised when a trailing window is instantiated with a lookback greater than
    the length of the underlying array.
    """

    msg = (
        "Can't construct a rolling window of length "
        "{window_length} on an array of length {nrows}."
    ).strip()


class WindowLengthNotPositive(ZiplineError):
    """Raised when a trailing window would be instantiated with a length less than
    1.
    """

    msg = ("Expected a window_length greater than 0, got {window_length}.").strip()


class NonWindowSafeInput(ZiplineError):
    """Raised when a Pipeline API term that is not deemed window safe is specified
    as an input to another windowed term.

    This is an error because it's generally not safe to compose windowed
    functions on split/dividend adjusted data.
    """

    msg = "Can't compute windowed expression {parent} with " "windowed input {child}."


class TermInputsNotSpecified(ZiplineError):
    """Raised if a user attempts to construct a term without specifying inputs and
    that term does not have class-level default inputs.
    """

    msg = "{termname} requires inputs, but no inputs list was passed."


class NonPipelineInputs(ZiplineError):
    """Raised when a non-pipeline object is passed as input to a ComputableTerm"""

    def __init__(self, term, inputs):
        self.term = term
        self.inputs = inputs

    def __str__(self):
        return (
            "Unexpected input types in {}. "
            "Inputs to Pipeline expressions must be Filters, Factors, "
            "Classifiers, or BoundColumns.\n"
            "Got the following type(s) instead: {}".format(
                type(self.term).__name__,
                sorted(set(map(type, self.inputs)), key=lambda t: t.__name__),
            )
        )


class TermOutputsEmpty(ZiplineError):
    """Raised if a user attempts to construct a term with an empty outputs list."""

    msg = "{termname} requires at least one output when passed an outputs " "argument."


class InvalidOutputName(ZiplineError):
    """Raised if a term's output names conflict with any of its attributes."""

    msg = (
        "{output_name!r} cannot be used as an output name for {termname}. "
        "Output names cannot start with an underscore or be contained in the "
        "following list: {disallowed_names}."
    )


class WindowLengthNotSpecified(ZiplineError):
    """Raised if a user attempts to construct a term without specifying window
    length and that term does not have a class-level default window length.
    """

    msg = "{termname} requires a window_length, but no window_length was passed."


class InvalidTermParams(ZiplineError):
    """Raised if a user attempts to construct a Term using ParameterizedTermMixin
    without specifying a `params` list in the class body.
    """

    msg = (
        "Expected a list of strings as a class-level attribute for "
        "{termname}.params, but got {value} instead."
    )


class DTypeNotSpecified(ZiplineError):
    """Raised if a user attempts to construct a term without specifying dtype and
    that term does not have class-level default dtype.
    """

    msg = "{termname} requires a dtype, but no dtype was passed."


class NotDType(ZiplineError):
    """Raised when a pipeline Term is constructed with a dtype that isn't a numpy
    dtype object.
    """

    msg = (
        "{termname} expected a numpy dtype "
        "object for a dtype, but got {dtype} instead."
    )


class UnsupportedDType(ZiplineError):
    """Raised when a pipeline Term is constructed with a dtype that's not
    supported.
    """

    msg = (
        "Failed to construct {termname}.\n"
        "Pipeline terms of dtype {dtype} are not yet supported."
    )


class BadPercentileBounds(ZiplineError):
    """Raised by API functions accepting percentile bounds when the passed bounds
    are invalid.
    """

    msg = (
        "Percentile bounds must fall between 0.0 and {upper_bound}, and min "
        "must be less than max."
        "\nInputs were min={min_percentile}, max={max_percentile}."
    )


class UnknownRankMethod(ZiplineError):
    """Raised during construction of a Rank factor when supplied a bad Rank
    method.
    """

    msg = "Unknown ranking method: '{method}'. " "`method` must be one of {choices}"


class AttachPipelineAfterInitialize(ZiplineError):
    """Raised when a user tries to call add_pipeline outside of initialize."""

    msg = (
        "Attempted to attach a pipeline after initialize(). "
        "attach_pipeline() can only be called during initialize."
    )


class PipelineOutputDuringInitialize(ZiplineError):
    """Raised when a user tries to call `pipeline_output` during initialize."""

    msg = (
        "Attempted to call pipeline_output() during initialize. "
        "pipeline_output() can only be called once initialize has completed."
    )


class NoSuchPipeline(ZiplineError, KeyError):
    """Raised when a user tries to access a non-existent pipeline by name."""

    msg = (
        "No pipeline named '{name}' exists. Valid pipeline names are {valid}. "
        "Did you forget to call attach_pipeline()?"
    )


class DuplicatePipelineName(ZiplineError):
    """Raised when a user tries to attach a pipeline with a name that already
    exists for another attached pipeline.
    """

    msg = (
        "Attempted to attach pipeline named {name!r}, but the name already "
        "exists for another pipeline. Please use a different name for this "
        "pipeline."
    )


class UnsupportedDataType(ZiplineError):
    """
    Raised by CustomFactors with unsupported dtypes.
    """

    def __init__(self, hint="", **kwargs):
        if hint:
            hint = " " + hint
        kwargs["hint"] = hint
        super(UnsupportedDataType, self).__init__(**kwargs)

    msg = "{typename} instances with dtype {dtype} are not supported.{hint}"


class NoFurtherDataError(ZiplineError):
    """Raised by calendar operations that would ask for dates beyond the extent of
    our known data.
    """

    # This accepts an arbitrary message string because it's used in more places
    # that can be usefully templated.
    msg = "{msg}"

    @classmethod
    def from_lookback_window(
            cls, initial_message, first_date, lookback_start, lookback_length
    ):
        return cls(
            msg=dedent(
                """
                {initial_message}

                lookback window started at {lookback_start}
                earliest known date was {first_date}
                {lookback_length} extra rows of data were required
                """
            ).format(
                initial_message=initial_message,
                first_date=first_date,
                lookback_start=lookback_start,
                lookback_length=lookback_length,
            )
        )


class UnsupportedDatetimeFormat(ZiplineError):
    """Raised when an unsupported datetime is passed to an API method."""

    msg = (
        "The input '{input}' passed to '{method}' is not "
        "coercible to a pandas.Timestamp object."
    )


class AssetDBVersionError(ZiplineError):
    """
    Raised by an AssetDBWriter or AssetFinder if the version number in the
    versions table does not match the ASSET_DB_VERSION in asset_writer.py.
    """

    msg = (
        "The existing Asset database has an incorrect version: {db_version}. "
        "Expected version: {expected_version}. Try rebuilding your asset "
        "database or updating your version of Zipline."
    )


class AssetDBImpossibleDowngrade(ZiplineError):
    msg = (
        "The existing Asset database is version: {db_version} which is lower "
        "than the desired downgrade version: {desired_version}."
    )


class HistoryWindowStartsBeforeData(ZiplineError):
    msg = (
        "History window extends before {first_trading_day}. To use this "
        "history window, start the backtest on or after {suggested_start_day}."
    )


class NonExistentAssetInTimeFrame(ZiplineError):
    msg = (
        "The target asset '{asset}' does not exist for the entire timeframe "
        "between {start_date} and {end_date}."
    )


class InvalidCalendarName(ZiplineError):
    """Raised when a calendar with an invalid name is requested."""

    msg = "The requested TradingCalendar, {calendar_name}, does not exist."


class CalendarNameCollision(ZiplineError):
    """
    Raised when the static calendar registry already has a calendar with a
    given name.
    """

    msg = "A calendar with the name {calendar_name} is already registered."


class CyclicCalendarAlias(ZiplineError):
    """
    Raised when calendar aliases form a cycle.
    """

    msg = "Cycle in calendar aliases: [{cycle}]"


class ScheduleFunctionWithoutCalendar(ZiplineError):
    """
    Raised when schedule_function is called but there is not a calendar to be
    used in the construction of an event rule.
    """

    # TODO update message when new TradingSchedules are built
    msg = (
        "To use schedule_function, the TradingAlgorithm must be running on an "
        "ExchangeTradingSchedule, rather than {schedule}."
    )


class ScheduleFunctionOutsideTradingStart(ZiplineError):
    """
    Raised when an algorithm schedules functions outside of
    before_trading_start()
    """
    msg = "schedule_function() should only be called in before_trading_start()"


class ScheduleFunctionInvalidCalendar(ZiplineError):
    """
    Raised when schedule_function is called with an invalid calendar argument.
    """

    msg = (
        "Invalid calendar '{given_calendar}' passed to schedule_function. "
        "Allowed options are {allowed_calendars}."
    )


class UnsupportedPipelineOutput(ZiplineError):
    """
    Raised when a 1D term is added as a column to a pipeline.
    """

    msg = (
        "Cannot add column {column_name!r} with term {term}. Adding slices or "
        "single-column-output terms as pipeline columns is not currently "
        "supported."
    )


class NonSliceableTerm(ZiplineError):
    """
    Raised when attempting to index into a non-sliceable term, e.g. instances
    of `ziplime.pipeline.term.LoadableTerm`.
    """

    msg = "Taking slices of {term} is not currently supported."


class IncompatibleTerms(ZiplineError):
    """
    Raised when trying to compute correlations/regressions between two 2D
    factors with different masks.
    """

    msg = (
        "{term_1} and {term_2} must have the same mask in order to compute "
        "correlations and regressions asset-wise."
    )


class LiquidityExceeded(Exception):
    pass


class NoDataOnDate(Exception):
    """
    Raised when a spot price cannot be found for the sid and date.
    """

    pass


class NotAssetConvertible(ValueError):
    pass


@dataclasses.dataclass
class BarSimulationError:
    message: str
    trace: str
    simulation_dt: datetime.datetime



import datetime
from abc import abstractmethod

from sqlalchemy.orm import Mapped, relationship, declared_attr

from ziplime.core.db.annotated_types import AssetRouterFKPK
from ziplime.core.db.base_model import BaseModel


class AssetModel(BaseModel):
    __tablename__ = "assets"
    __abstract__ = True

    sid: Mapped[AssetRouterFKPK]
    mic: Mapped[str | None]
    # asset_router = relationship(AssetRouter)
    asset_name: Mapped[str]
    start_date: Mapped[datetime.date]
    end_date: Mapped[datetime.date]
    first_traded: Mapped[datetime.date]
    auto_close_date: Mapped[datetime.date]

    @declared_attr  # type: ignore [misc]
    def asset_router(cls):
        return relationship("AssetRouter", foreign_keys=f"{cls.__name__}.sid")

    def __hash__(self):
        return self.sid

    @abstractmethod
    def get_symbol_by_exchange(self, exchange_name: str | None) -> str | None: ...



import datetime
from abc import abstractmethod

import structlog
from exchange_calendars import ExchangeCalendar


class TradingClock:
    trading_calendar: ExchangeCalendar
    emission_rate: datetime.timedelta

    def __init__(self, trading_calendar: ExchangeCalendar, emission_rate: datetime.timedelta):
        self.trading_calendar = trading_calendar
        self.emission_rate = emission_rate
        self._logger = structlog.get_logger(__name__)

    @abstractmethod
    def __iter__(self):
        raise NotImplementedError("__iter__ method must be implemented for trading clock.")



from dataclasses import dataclass

from ziplime.assets.entities.asset import Asset
from ziplime.assets.entities.equity_symbol_mapping import EquitySymbolMapping


@dataclass(frozen=True)
class Equity(Asset):
    # mapping exchange code -> symbol
    symbol_mapping: dict[str, EquitySymbolMapping]

    def get_symbol_by_exchange(self, exchange_name: str | None) -> str | None:
        if exchange_name is None:
            symbol_mapping = next(iter(self.symbol_mapping.values()), None)
        else:
            symbol_mapping = self.symbol_mapping.get(exchange_name, None)
        if symbol_mapping:
            return symbol_mapping.symbol
        return None

    def __hash__(self):
        return hash(self.sid)



from sqlalchemy.orm import Mapped

from ziplime.core.db.annotated_types import IntegerPK
from ziplime.core.db.base_model import BaseModel


class AssetRouter(BaseModel):
    __tablename__ = "asset_router"

    sid: Mapped[IntegerPK]
    asset_type: Mapped[str]



import datetime

from toolz import partition_all

from ziplime.constants.period import Period


def compute_date_range_chunks(sessions, start_date, end_date, chunksize):
    """Compute the start and end dates to run a pipeline for.

    Parameters
    ----------
    sessions : DatetimeIndex
        The available dates.
    start_date : pd.Timestamp
        The first date in the pipeline.
    end_date : pd.Timestamp
        The last date in the pipeline.
    chunksize : int or None
        The size of the chunks to run. Setting this to None returns one chunk.

    Returns
    -------
    ranges : iterable[(np.datetime64, np.datetime64)]
        A sequence of start and end dates to run the pipeline for.
    """
    if start_date not in sessions:
        raise KeyError(
            "Start date %s is not found in calendar."
            % (start_date.strftime("%Y-%m-%d"),)
        )
    if end_date not in sessions:
        raise KeyError(
            "End date %s is not found in calendar." % (end_date.strftime("%Y-%m-%d"),)
        )
    if end_date < start_date:
        raise ValueError(
            "End date %s cannot precede start date %s."
            % (end_date.strftime("%Y-%m-%d"), start_date.strftime("%Y-%m-%d"))
        )

    if chunksize is None:
        return [(start_date, end_date)]

    start_ix, end_ix = sessions.slice_locs(start_date, end_date)
    return ((r[0], r[-1]) for r in partition_all(chunksize, sessions[start_ix:end_ix]))


def make_utc_aware(dti):
    """Normalizes a pd.DateTimeIndex. Assumes UTC if tz-naive."""
    try:
        # ensure tz-aware Timestamp has tz UTC
        return dti.tz_convert(tz="UTC")
    except TypeError:
        # if naive, instead convert timestamp to UTC
        return dti.tz_localize(tz="UTC")


def strip_time_and_timezone_info(dt: datetime.datetime) -> datetime.datetime:
    return dt.replace(hour=0, minute=0, second=0, microsecond=0)


def period_to_timedelta(period: Period | datetime.timedelta) -> datetime.timedelta:
    if type(period) is datetime.timedelta:
        return period
    if period == "1us":
        return datetime.timedelta(microseconds=1)
    if period == "1ms":
        return datetime.timedelta(milliseconds=1)
    if period == "1s":
        return datetime.timedelta(seconds=1)
    if period == "1m":
        return datetime.timedelta(minutes=1)
    if period == "1h":
        return datetime.timedelta(hours=1)
    if period == "1d":
        return datetime.timedelta(days=1)
    if period == "1w":
        return datetime.timedelta(days=7)
    if period == "1mo":
        return datetime.timedelta(days=30)
    if period == "1q":
        return datetime.timedelta(days=90)
    if period == "1y":
        return datetime.timedelta(days=365)

    raise ValueError(f"Invalid period: {period}")



import asyncio
import datetime
import logging
import os
from pathlib import Path

import asyncclick as click

from ziplime.assets.domain.ordered_contracts import CHAIN_PREDICATES
from ziplime.assets.repositories.sqlalchemy_adjustments_repository import SqlAlchemyAdjustmentRepository
from ziplime.assets.repositories.sqlalchemy_asset_repository import SqlAlchemyAssetRepository
from ziplime.constants.fundamental_data import FUNDAMENTAL_DATA_COLUMNS
from ziplime.data.services.bundle_service import BundleService
from ziplime.data.services.file_system_bundle_registry import FileSystemBundleRegistry
from ziplime.data.services.limex_hub_data_source import LimexHubDataSource
from ziplime.data.services.file_system_parquet_bundle_storage import FileSystemParquetBundleStorage
from ziplime.domain.benchmark_spec import BenchmarkSpec
from ziplime.domain.data_frequency import DataFrequency
from ziplime.finance.commission import DEFAULT_MINIMUM_COST_PER_FUTURE_TRADE, DEFAULT_PER_CONTRACT_COST, PerContract, \
    DEFAULT_MINIMUM_COST_PER_EQUITY_TRADE, DEFAULT_PER_SHARE_COST, PerShare
from ziplime.finance.constants import FUTURE_EXCHANGE_FEES_BY_SYMBOL
from ziplime.finance.domain.simulation_paremeters import SimulationParameters
from ziplime.finance.metrics import default_metrics
from ziplime.finance.slippage.fixed_basis_points_slippage import FixedBasisPointsSlippage
from ziplime.finance.slippage.slippage_model import DEFAULT_FUTURE_VOLUME_SLIPPAGE_BAR_LIMIT
from ziplime.finance.slippage.volatility_volume_share import VolatilityVolumeShare
from ziplime.gens.domain.simulation_clock import SimulationClock
from ziplime.exchanges.lime_trader_sdk.lime_trader_sdk_exchange import LimeTraderSdkExchange
from ziplime.exchanges.simulation_exchange import SimulationExchange
from ziplime.utils.date_utils import strip_time_and_timezone_info
from ziplime.utils.run_algo import run_algorithm
from exchange_calendars import get_calendar as ec_get_calendar

from asyncclick import DateTime
from ziplime.utils.cli import Timestamp

from ziplime.utils.bundle_utils import get_fundamental_data_provider, get_data_source


def validate_date_range(date_min: datetime.datetime, date_max: datetime.datetime):
    def _validate_date_range(ctx, param, value):
        if not date_min < value.replace(tzinfo=datetime.timezone.utc) < date_max:
            raise click.BadParameter(f"Must be between {date_min} and {date_max}")
        return value

    return _validate_date_range


@click.group()
@click.pass_context
async def main(ctx):
    """Top level ziplime entry point."""
    # install a logging handler before performing any other operations
    logging.basicConfig(
        format="[%(asctime)s-%(levelname)s][%(name)s]\n %(message)s",
        level=logging.INFO,
        datefmt="%Y-%m-%dT%H:%M:%S%z",
    )


@main.command(context_settings=dict(
))
@click.option(
    "-b",
    "--bundle",
    help="The data bundle to ingest.",
)
@click.option(
    "-c",
    "--trading-calendar",
    help="Default calendar to use.",
)
@click.option(
    "--start-date",
    type=DateTime(),
    callback=validate_date_range(date_min=datetime.datetime(year=1800,
                                                            month=1,
                                                            day=1,
                                                            tzinfo=datetime.timezone.utc
                                                            ),
                                 date_max=datetime.datetime.now(tz=datetime.timezone.utc),
                                 )
)
@click.option(
    "--end-date",
    type=DateTime(),
    callback=validate_date_range(date_min=datetime.datetime(year=1800,
                                                            month=1,
                                                            day=1,
                                                            tzinfo=datetime.timezone.utc
                                                            ),
                                 date_max=datetime.datetime.now(tz=datetime.timezone.utc),
                                 )
)
@click.option(
    "--frequency",
    default="1d",
    type=click.Choice([df.value for df in DataFrequency]),
)
@click.option("-s", '--symbols')
@click.option('--fundamental-data')
@click.option(
    "--show-progress/--no-show-progress",
    default=True,
    help="Print progress information to the terminal.",
)
@click.option(
    "--historical-market-data-provider",
    type=click.Choice(['limex-hub', 'lime-trader-sdk']),
    default="limex-hub",
    help="Market data provider for historical data",
    show_default=True,
)
@click.option(
    "--fundamental-data-provider",
    type=click.Choice(['limex-hub', ]),
    default="limex-hub",
    help="Fundamental data provider",
    show_default=True,
)
@click.option(
    "--skip-fundamental-data",
    default=False,
    is_flag=True,
    help="If passed, fundamental data won't be ingested.",
)
@click.option(
    "--bundle-storage-path",
    default=Path(Path.home(), ".ziplime", "data"),
    show_default=True,
    help="Path to the bundle storage on filesystem.",
)
@click.pass_context
async def ingest(ctx, bundle, start_date, end_date, frequency, symbols, fundamental_data, show_progress,
                 trading_calendar,
                 historical_market_data_provider,
                 fundamental_data_provider,
                 skip_fundamental_data,
                 bundle_storage_path,
                 ):
    """Top level ziplime entry point."""
    symbols_parsed = symbols.split(',') if symbols else None
    if skip_fundamental_data:
        fundamental_data_list = set()
    else:
        fundamental_data_list = fundamental_data.split(",") if fundamental_data else None
    # install a logging handler before performing any other operations
    fundamental_data_list_cols = fundamental_data_list if fundamental_data_list is not None else [
        col.name for col in
        FUNDAMENTAL_DATA_COLUMNS
    ]
    fundamental_data_provider_instance = get_fundamental_data_provider(code=fundamental_data_provider)
    fundamental_data_column_names = fundamental_data_provider_instance.get_fundamental_data_column_names(
        fundamental_data_fields=set(fundamental_data_list_cols))
    fundamental_data_cols = (
        [
            col
            for col in FUNDAMENTAL_DATA_COLUMNS
            if col.name in fundamental_data_column_names
        ]
        if fundamental_data_list is not None
        else FUNDAMENTAL_DATA_COLUMNS
    )
    bundle_registry = FileSystemBundleRegistry(base_data_path=bundle_storage_path)
    bundle_service = BundleService(bundle_registry=bundle_registry)
    bundle_storage = FileSystemParquetBundleStorage(base_data_path=bundle_storage_path, compression_level=5)
    data_bundle_source = LimexHubDataSource.from_env()

    calendar = ec_get_calendar(trading_calendar, start=start_date - datetime.timedelta(days=30))

    bundle_version = str(int(datetime.datetime.now(tz=calendar.tz).timestamp()))
    assets_repository = SqlAlchemyAssetRepository(base_storage_path=bundle_storage_path,
                                                  bundle_name=bundle,
                                                  bundle_version=bundle_version,
                                                  future_chain_predicates=CHAIN_PREDICATES)
    adjustments_repository = SqlAlchemyAdjustmentRepository(base_storage_path=bundle_storage_path,
                                                            bundle_name=bundle,
                                                            bundle_version=bundle_version)

    await bundle_service.ingest_bundle(
        date_start=start_date.replace(tzinfo=calendar.tz),
        date_end=end_date.replace(tzinfo=calendar.tz),
        bundle_storage=bundle_storage,
        data_bundle_source=data_bundle_source,
        frequency=DataFrequency(frequency).to_timedelta(),
        symbols=symbols_parsed,
        name=bundle,
        bundle_version=bundle_version,
        trading_calendar=calendar,
        assets_repository=assets_repository,
        adjustments_repository=adjustments_repository,
        forward_fill_missing_ohlcv_data=True
    )


@main.command(context_settings=dict(
    ignore_unknown_options=True,
    allow_extra_args=True,
))
@click.option(
    "--bundle-storage-path",
    default=Path(Path.home(), ".ziplime", "data"),
    show_default=True,
    help="Path to the bundle storage on filesystem.",
)
@click.option(
    "-b",
    "--bundle",
    metavar="BUNDLE-NAME",
    show_default=True,
    help="The data bundle to clean.",
)
@click.option(
    "-e",
    "--before",
    type=Timestamp(),
    help="Clear all data before TIMESTAMP."
         " This may not be passed with -k / --keep-last",
)
@click.option(
    "-a",
    "--after",
    type=Timestamp(),
    help="Clear all data after TIMESTAMP"
         " This may not be passed with -k / --keep-last",
)
@click.option(
    "-k",
    "--keep-last",
    type=int,
    metavar="N",
    help="Clear all but the last N downloads."
         " This may not be passed with -e / --before or -a / --after",
)
@click.pass_context
async def clean(ctx, bundle_storage_path, bundle, before, after, keep_last):
    """Top level ziplime entry point."""

    bundle_registry = FileSystemBundleRegistry(base_data_path=bundle_storage_path)

    bundle_service = BundleService(bundle_registry=bundle_registry)
    await bundle_service.clean(bundle_name=bundle, after=after, before=before,
                               keep_last=keep_last)


@main.command(context_settings=dict(
    ignore_unknown_options=True,
    allow_extra_args=True,
))
@click.option(
    "--bundle-storage-path",
    default=Path(Path.home(), ".ziplime", "data"),
    show_default=True,
    help="Path to the bundle storage on filesystem.",
)
@click.pass_context
async def bundles(ctx, bundle_storage_path):
    """Top level ziplime entry point."""
    bundle_registry = FileSystemBundleRegistry(base_data_path=bundle_storage_path)

    for bundle in await bundle_registry.list_bundles():
        click.echo(f"{bundle["name"]} {bundle["version"]} - {bundle['timestamp']}")


@main.command()
@click.option(
    "-f",
    "--algofile",
    default=None,
    type=click.File("r"),
    help="The file that contains the algorithm to run.",
)
@click.option(
    "--emission-rate",
    type=click.Choice([df.value for df in DataFrequency]),
    show_default=True,
    help="Emission rate of the simulation.",
)
@click.option(
    "--capital-base",
    type=float,
    help="The starting capital for the simulation.",
)
@click.option(
    "-b",
    "--bundle",
    metavar="BUNDLE-NAME",
    help="The data bundle to use for the simulation.",
)
@click.option(
    "-bf",
    "--benchmark-file",
    default=None,
    type=click.Path(exists=True, dir_okay=False, readable=True, path_type=str),
    help="The csv file that contains the benchmark returns",
)
@click.option(
    "--benchmark-symbol",
    default=None,
    type=click.STRING,
    help="The symbol of the instrument to be used as a benchmark "
         "(should exist in the ingested bundle)",
)
@click.option(
    "--benchmark-sid",
    default=None,
    type=int,
    help="The sid of the instrument to be used as a benchmark "
         "(should exist in the ingested bundle)",
)
@click.option(
    "--no-benchmark",
    is_flag=True,
    default=False,
    help="If passed, use a benchmark of zero returns.",
)
@click.option(
    "-s",
    "--start-date",
    type=click.DateTime(),
    help="The start date of the simulation.",
)
@click.option(
    "-e",
    "--end-date",
    type=click.DateTime(),
    help="The end date of the simulation.",
)
@click.option(
    "-o",
    "--output",
    default="-",
    metavar="FILENAME",
    show_default=True,
    help="The location to write the perf data. If this is '-' the perf will"
         " be written to stdout.",
)
@click.option(
    "--trading-calendar",
    metavar="TRADING-CALENDAR",
    default="NYSE",
    help="The calendar you want to use e.g. XLON. XNYS is the default.",
)
@click.option(
    "--print-algo/--no-print-algo",
    is_flag=True,
    default=False,
    help="Print the algorithm to stdout.",
)
@click.option(
    "--exchange-name",
    default='LIME',
    help="The exchange to use for trading.",
    show_default=True,
)
@click.option(
    "--exchange-type",
    default='simulation',
    type=click.Choice(['simulation', 'lime-trader-sdk']),
    help="The exchange to use for trading.",
    show_default=True,
)
@click.option(
    "--live-market-data-provider",
    type=click.Choice(['lime-trader-sdk']),
    default=None,
    help="Market data provider for live trading",
)
@click.option(
    "--bundle-storage-path",
    default=Path(Path.home(), ".ziplime", "data"),
    show_default=True,
    help="Path to the bundle storage on filesystem.",
)
@click.pass_context
async def run(
        ctx,
        algofile,
        emission_rate,
        capital_base,
        bundle,
        benchmark_file,
        benchmark_symbol,
        benchmark_sid,
        no_benchmark,
        start_date,
        end_date,
        output,
        trading_calendar,
        print_algo,
        bundle_storage_path,
        exchange_name: str,
        exchange_type: str,
        live_market_data_provider: str | None,
):
    """Run a backtest for the given algorithm."""

    calendar = ec_get_calendar(trading_calendar,
                               start=strip_time_and_timezone_info(start_date) - datetime.timedelta(days=30))

    benchmark_spec = BenchmarkSpec(
        benchmark_returns=None,
        benchmark_sid=benchmark_sid,
        benchmark_symbol=benchmark_symbol,
        benchmark_file=benchmark_file,
        no_benchmark=no_benchmark,
    )
    algotext = None
    if algofile is not None:
        algotext = algofile.read()
    bundle_registry = FileSystemBundleRegistry(base_data_path=bundle_storage_path)

    if exchange_type == "simulation":
        exchange_class = SimulationExchange(
            name=exchange_name,
            equity_slippage=FixedBasisPointsSlippage(),
            equity_commission=PerShare(
                cost=DEFAULT_PER_SHARE_COST,
                min_trade_cost=DEFAULT_MINIMUM_COST_PER_EQUITY_TRADE,

            ),
            future_slippage=VolatilityVolumeShare(
                volume_limit=DEFAULT_FUTURE_VOLUME_SLIPPAGE_BAR_LIMIT,
            ),
            future_commission=PerContract(
                cost=DEFAULT_PER_CONTRACT_COST,
                exchange_fee=FUTURE_EXCHANGE_FEES_BY_SYMBOL,
                min_trade_cost=DEFAULT_MINIMUM_COST_PER_FUTURE_TRADE
            ),
        )
    elif exchange_type == "lime-trader-sdk":
        exchange_class = LimeTraderSdkExchange(
            name=exchange_name,
            lime_sdk_credentials_file=None
        )

    else:
        raise Exception("Not valid exchange.")
    max_shares = int(1e11)

    sim_params = SimulationParameters(
        start_date=start_date.replace(tzinfo=calendar.tz),
        end_date=end_date.replace(tzinfo=calendar.tz),
        trading_calendar=calendar,
        capital_base=capital_base,
        emission_rate=DataFrequency(emission_rate).to_timedelta(),
        max_shares=max_shares,
        exchange=exchange_class,
        bundle_name=bundle
    )

    clock = SimulationClock(
        sessions=sim_params.sessions,
        market_opens=sim_params.market_opens,
        market_closes=sim_params.market_closes,
        before_trading_start_minutes=sim_params.before_trading_start_minutes,
        emission_rate=sim_params.emission_rate,
        timezone=sim_params.trading_calendar.tz
    )
    timedelta_diff_from_current_time = datetime.datetime.now(tz=sim_params.trading_calendar.tz) - start_date.replace(
        tzinfo=sim_params.trading_calendar.tz)

    # clock = RealtimeClock(
    #     sessions=sim_params.sessions,
    #     market_opens=sim_params.market_opens,
    #     market_closes=sim_params.market_closes,
    #     before_trading_start_minutes=sim_params.before_trading_start_minutes,
    #     emission_rate=sim_params.emission_rate,
    #     timezone=sim_params.trading_calendar.tz,
    #     timedelta_diff_from_current_time=-timedelta_diff_from_current_time
    # )

    result = await run_algorithm(
        algofile=getattr(algofile, "name", "<algorithm>"),
        algotext=algotext,
        print_algo=print_algo,
        metrics_set=default_metrics(),
        benchmark_spec=benchmark_spec,
        custom_loader=None,
        missing_data_bundle_source=get_data_source(
            live_market_data_provider) if live_market_data_provider is not None else None,
        bundle_registry=bundle_registry,
        simulation_params=sim_params,
        clock=clock

    )

    if output == "-":
        click.echo(str(result))
    elif output != os.devnull:  # make the ziplime magic not write any data
        # TODO: test this
        result.to_pickle(output)
    return result


if __name__ == "__main__":
    asyncio.run(main())



import datetime

import polars as pl
from exchange_calendars import ExchangeCalendar

from ziplime.trading.enums.simulation_event import SimulationEvent
from ziplime.gens.domain.trading_clock import TradingClock


class SimulationClock(TradingClock):
    def __init__(self,
                 start_date: datetime.datetime,
                 end_date: datetime.datetime,
                 trading_calendar: ExchangeCalendar,
                 emission_rate: datetime.timedelta):
        super().__init__(trading_calendar=trading_calendar, emission_rate=emission_rate)
        if start_date >= end_date:
            raise ValueError("Period start falls after period end.")
        if start_date >= trading_calendar.last_session.replace(tzinfo=trading_calendar.tz):
            raise ValueError("Period start falls after the last known trading day.")
        if end_date <= trading_calendar.first_session.replace(tzinfo=trading_calendar.tz):
            raise ValueError("Period end falls before the first known trading day.")

            # chop off any minutes or hours on the given start and end dates,
            # as we only support session labels here (and we represent session
            # labels as midnight UTC).

        self.start_date = start_date
        self.end_date = end_date

        self.start_session = start_date.date()
        self.end_session = end_date.date()

        if trading_calendar.sessions_distance(self.start_session, self.end_session) < 1:
            raise Exception(
                f"There are no trading days between {self.start_session} and {self.end_session}")
        if not trading_calendar.is_session(self.start_session):
            # if the start date is not a valid session in this calendar,
            # push it forward to the first valid session
            self.start_session = trading_calendar.minute_to_session(
                self.start_session
            ).tz_localize(self.trading_calendar.tz).to_pydatetime().date()

        if not trading_calendar.is_session(self.end_session):
            # if the end date is not a valid session in this calendar,
            # pull it backward to the last valid session before the given
            # end date.
            self.end_session = trading_calendar.minute_to_session(
                self.end_session, direction="previous"
            ).tz_localize(self.trading_calendar.tz).to_pydatetime().date()

        self.first_open = trading_calendar.session_first_minute(
            self.start_session
        ).tz_convert(self.trading_calendar.tz).to_pydatetime()
        self.last_close = trading_calendar.session_close(
            self.end_session
        ).tz_convert(self.trading_calendar.tz).to_pydatetime()

        self.sessions = pl.Series(self.trading_calendar.sessions_in_range(
            self.start_session, self.end_session)
        ).dt.date()

        self.market_closes = pl.Series(
            self.trading_calendar.schedule.loc[self.sessions, "close"].dt.tz_convert(
                self.trading_calendar.tz))
        self.market_opens = pl.Series(
            self.trading_calendar.first_minutes.loc[self.sessions].dt.tz_convert(
                self.trading_calendar.tz))

        self.before_trading_start_minutes = self.market_opens - datetime.timedelta(minutes=46)
        self.minutes_by_session = self.calc_minutes_by_session()

    def calc_minutes_by_session(self):
        minutes_by_session_n = {}
        if self.emission_rate < datetime.timedelta(days=1):
            for session_idx, session in enumerate(self.sessions):
                minutes = pl.datetime_range(self.market_opens[session_idx], self.market_closes[session_idx],
                                            interval=self.emission_rate,
                                            eager=True)
                minutes_by_session_n[session] = minutes
        else:
            minutes_by_session_n = {session: pl.Series([self.market_closes[session_idx]]) for
                                    session_idx, session in enumerate(self.sessions)}
        return minutes_by_session_n

    def __iter__(self):
        for idx, session in enumerate(self.sessions):
            yield session, SimulationEvent.SESSION_START

            bts_minute = self.before_trading_start_minutes[idx]
            regular_minutes = self.minutes_by_session[session]

            yield bts_minute, SimulationEvent.BEFORE_TRADING_START_BAR
            for minute in regular_minutes:
                yield minute, SimulationEvent.BAR
                yield minute, SimulationEvent.EMISSION_RATE_END

            yield regular_minutes[-1], SimulationEvent.SESSION_END



"""
Mixins classes for use with Filters and Factors.

The mixin classes inherit from Term to ensure they appear before
Term in the MRO of any class using the mixin
"""

from abc import abstractmethod

from numpy import (
    array,
    full,
    recarray,
    searchsorted,
    vstack,
    where,
)
from pandas import NaT as pd_NaT

from ziplime.errors import (
    WindowLengthNotPositive,
    UnsupportedDataType,
    NonExistentAssetInTimeFrame,
    NoFurtherDataError,
)
from ziplime.lib.labelarray import LabelArray, labelarray_where
from ziplime.utils.context_tricks import nop_context
from ziplime.utils.numpy_utils import bool_dtype
from ziplime.utils.pandas_utils import nearest_unequal_elements

from .downsample_helpers import (
    select_sampling_indices, SUPPORTED_DOWNSAMPLE_FREQUENCIES,
)
from .term import Term


class PositiveWindowLengthMixin(Term):
    """
    Validation mixin enforcing that a Term gets a positive WindowLength
    """

    def _validate(self):
        super(PositiveWindowLengthMixin, self)._validate()
        if not self.windowed:
            raise WindowLengthNotPositive(window_length=self.window_length)


class SingleInputMixin(Term):
    """
    Validation mixin enforcing that a Term gets a length-1 inputs list.
    """

    def _validate(self):
        super(SingleInputMixin, self)._validate()
        num_inputs = len(self.inputs)
        if num_inputs != 1:
            raise ValueError(
                "{typename} expects only one input, "
                "but received {num_inputs} instead.".format(
                    typename=type(self).__name__, num_inputs=num_inputs
                )
            )


class StandardOutputs(Term):
    """
    Validation mixin enforcing that a Term cannot produce non-standard outputs.
    """

    def _validate(self):
        super(StandardOutputs, self)._validate()
        if self.outputs is not None:
            raise ValueError(
                "{typename} does not support custom outputs,"
                " but received custom outputs={outputs}.".format(
                    typename=type(self).__name__,
                    outputs=self.outputs,
                )
            )


class RestrictedDTypeMixin(Term):
    """
    Validation mixin enforcing that a term has a specific dtype.
    """

    ALLOWED_DTYPES = None

    def _validate(self):
        super(RestrictedDTypeMixin, self)._validate()
        assert self.ALLOWED_DTYPES is not None, (
                "ALLOWED_DTYPES not supplied on subclass "
                "of RestrictedDTypeMixin: %s." % type(self).__name__
        )

        if self.dtype not in self.ALLOWED_DTYPES:
            raise UnsupportedDataType(
                typename=type(self).__name__,
                dtype=self.dtype,
            )


class CustomTermMixin(Term):
    """
    Mixin for user-defined rolling-window Terms.

    Implements `_compute` in terms of a user-defined `compute` function, which
    is mapped over the input windows.

    Used by CustomFactor, CustomFilter, CustomClassifier, etc.
    """

    ctx = nop_context

    def __new__(
            cls,
            inputs=None,
            outputs=None,
            window_length=None,
            mask=None,
            dtype=None,
            missing_value=None,
            ndim=None,
            **kwargs,
    ):

        unexpected_keys = set(kwargs) - set(cls.params)
        if unexpected_keys:
            raise TypeError(
                "{termname} received unexpected keyword "
                "arguments {unexpected}".format(
                    termname=cls.__name__,
                    unexpected={k: kwargs[k] for k in unexpected_keys},
                )
            )

        return super(CustomTermMixin, cls).__new__(
            cls,
            inputs=inputs,
            outputs=outputs,
            window_length=window_length,
            mask=mask,
            dtype=dtype,
            missing_value=missing_value,
            ndim=ndim,
            **kwargs,
        )

    def compute(self, today, assets, out, *arrays):
        """
        Override this method with a function that writes a value into `out`.
        """
        raise NotImplementedError(
            "{name} must define a compute method".format(name=type(self).__name__)
        )

    def _allocate_output(self, windows, shape):
        """
        Allocate an output array whose rows should be passed to `self.compute`.

        The resulting array must have a shape of ``shape``.

        If we have standard outputs (i.e. self.outputs is None), the
        default is an empty ndarray whose dtype is ``self.dtype``.

        If we have an outputs tuple, the default is an empty recarray with
        ``self.outputs`` as field names. Each field will have dtype
        ``self.dtype``.

        This can be overridden to control the kind of array constructed
        (e.g. to produce a LabelArray instead of an ndarray).
        """
        missing_value = self.missing_value
        outputs = self.outputs
        if outputs is not None:
            out = recarray(
                shape,
                formats=[self.dtype.str] * len(outputs),
                names=outputs,
            )
            out[:] = missing_value
        else:
            out = full(shape, missing_value, dtype=self.dtype)
        return out

    def _format_inputs(self, windows, column_mask):
        inputs = []
        for input_ in windows:
            window = next(input_)
            if window.shape[1] == 1:
                # Do not mask single-column inputs.
                inputs.append(window)
            else:
                inputs.append(window[:, column_mask])
        return inputs

    def _compute(self, windows, dates, assets, mask):
        """
        Call the user's `compute` function on each window with a pre-built
        output array.
        """
        format_inputs = self._format_inputs
        compute = self.compute
        params = self.params
        ndim = self.ndim

        shape = (len(mask), 1) if ndim == 1 else mask.shape
        out = self._allocate_output(windows, shape)

        with self.ctx:
            for idx, date in enumerate(dates):
                # Never apply a mask to 1D outputs.
                out_mask = array([True]) if ndim == 1 else mask[idx]

                # Mask our inputs as usual.
                inputs_mask = mask[idx]

                masked_assets = assets[inputs_mask]
                out_row = out[idx][out_mask]
                inputs = format_inputs(windows, inputs_mask)

                compute(date, masked_assets, out_row, *inputs, **params)
                out[idx][out_mask] = out_row
        return out

    def graph_repr(self):
        """Short repr to use when rendering Pipeline graphs."""
        # Graphviz interprets `\l` as "divide label into lines, left-justified"
        return type(self).__name__ + ":\\l  window_length: %d\\l" % self.window_length


class LatestMixin(SingleInputMixin):
    """
    Common behavior for :attr:`ziplime.pipeline.data.BoundColumn.latest`.

    Given a :class:`~ziplime.pipeline.data.DataSet` named ``MyData`` with a
    column ``col`` of numeric dtype, the following expression:

    .. code-block:: python

       factor = MyData.col.latest

    is equivalent to:

    .. code-block:: python

       class Latest(CustomFactor):
           inputs = [MyData.col]
           window_length = 1

           def compute(self, today, assets, out, data):
               out[:] = data[-1]

       factor = Latest()

    The behavior is the same for columns of boolean or string dtype, except the
    resulting expression will be a :class:`~ziplime.pipeline.CustomFilter` for
    boolean columns, and the resulting object will be a
    :class:`~ziplime.pipeline.CustomClassifier` for string or integer columns.
    """

    window_length = 1

    def compute(self, today, assets, out, data):
        out[:] = data[-1]

    def _validate(self):
        super(LatestMixin, self)._validate()
        if self.inputs[0].dtype != self.dtype:
            raise TypeError(
                "{name} expected an input of dtype {expected}, "
                "but got {actual} instead.".format(
                    name=type(self).__name__,
                    expected=self.dtype,
                    actual=self.inputs[0].dtype,
                )
            )

    def graph_repr(self):
        return "Latest"


class UniversalMixin(Term):
    """
    Base class for "universal" mixins.

    Universal mixins are used to implement expressions that need separate
    subclasses for each of the ComputableTerm subclasses (Factor, Filter, and
    Classifier). Such expressions are usually return types of methods of
    ComputableTerm, such as `downsample()`, `alias()`, or `fillna()`.

    A type may only inherit from one UniversalMixin.
    """

    # Memo dict mapping pairs of (mixin_type, principal_type) to subtypes.
    _UNIVERSAL_MIXIN_SUBTYPES = {}

    @staticmethod
    @abstractmethod
    def _universal_mixin_type():
        raise NotImplementedError("_universal_mixin_type")

    @staticmethod
    @abstractmethod
    def _universal_mixin_specialization_name(principal_type):
        raise NotImplementedError("_universal_mixin_specialization_name")

    @classmethod
    def universal_mixin_specialization(cls, principal_type):
        """
        Create a new subtype of `principal_type` that adds this mixin to
        ``principal_type``. ``principal_type`` will be one of Factor, Filter,
        or Classifier.
        """
        mixin = cls._universal_mixin_type()
        memo_key = (mixin, principal_type)

        try:
            return cls._UNIVERSAL_MIXIN_SUBTYPES[memo_key]
        except KeyError:
            new_type = type(
                mixin._universal_mixin_specialization_name(principal_type),
                (mixin, principal_type),
                {"__module__": principal_type.__module__},
            )
            cls._UNIVERSAL_MIXIN_SUBTYPES[memo_key] = new_type
            return new_type


class AliasedMixin(SingleInputMixin, UniversalMixin):
    """
    Mixin for aliased terms.
    """

    def __new__(cls, term, name):
        return super(AliasedMixin, cls).__new__(
            cls,
            inputs=(term,),
            outputs=term.outputs,
            window_length=0,
            name=name,
            dtype=term.dtype,
            missing_value=term.missing_value,
            ndim=term.ndim,
            window_safe=term.window_safe,
        )

    def _init(self, name, *args, **kwargs):
        self.name = name
        return super(AliasedMixin, self)._init(*args, **kwargs)

    @classmethod
    def _static_identity(cls, name, *args, **kwargs):
        return (
            super(AliasedMixin, cls)._static_identity(*args, **kwargs),
            name,
        )

    def _compute(self, inputs, dates, assets, mask):
        return inputs[0]

    def __repr__(self):
        return "{type}({inner}, name={name!r})".format(
            type=type(self).__name__,
            inner=self.inputs[0].recursive_repr(),
            name=self.name,
        )

    def graph_repr(self):
        """Short repr to use when rendering Pipeline graphs."""
        return self.name

    @staticmethod
    def _universal_mixin_type():
        return AliasedMixin

    @staticmethod
    def _universal_mixin_specialization_name(principal_type):
        return "Aliased" + principal_type.__name__


class DownsampledMixin(StandardOutputs, UniversalMixin):
    """Universal mixin for downsampled terms.

    A downsampled term is a wrapper around the "real" term that performs actual
    computation. The downsampler is responsible for calling the real term's
    `compute` method at selected intervals and forward-filling the computed
    values.

    Downsampling is not currently supported for terms with multiple outputs.
    """

    # There's no reason to take a window of a downsampled term.  The whole
    # point is that you're re-using the same result multiple times.
    window_safe = False

    def __new__(cls, term: Term, frequency):
        if frequency not in SUPPORTED_DOWNSAMPLE_FREQUENCIES:
            raise ValueError(
                "Downsampled terms only support frequencies in %s, but got %s."
                % (SUPPORTED_DOWNSAMPLE_FREQUENCIES, frequency)
            )
        return super(DownsampledMixin, cls).__new__(
            cls,
            inputs=term.inputs,
            outputs=term.outputs,
            window_length=term.window_length,
            mask=term.mask,
            frequency=frequency,
            wrapped_term=term,
            dtype=term.dtype,
            missing_value=term.missing_value,
            ndim=term.ndim,
        )

    def _init(self, frequency, wrapped_term, *args, **kwargs):
        self._frequency = frequency
        self._wrapped_term = wrapped_term
        return super(DownsampledMixin, self)._init(*args, **kwargs)

    @classmethod
    def _static_identity(cls, frequency, wrapped_term, *args, **kwargs):
        return (
            super(DownsampledMixin, cls)._static_identity(*args, **kwargs),
            frequency,
            wrapped_term,
        )

    def compute_extra_rows(self, all_dates, start_date, end_date, min_extra_rows):
        """
        Ensure that min_extra_rows pushes us back to a computation date.

        Parameters
        ----------
        all_dates : pd.DatetimeIndex
            The trading sessions against which ``self`` will be computed.
        start_date : pd.Timestamp
            The first date for which final output is requested.
        end_date : pd.Timestamp
            The last date for which final output is requested.
        min_extra_rows : int
            The minimum number of extra rows required of ``self``, as
            determined by other terms that depend on ``self``.

        Returns
        -------
        extra_rows : int
            The number of extra rows to compute.  This will be the minimum
            number of rows required to make our computed start_date fall on a
            recomputation date.
        """
        try:
            current_start_pos = all_dates.get_loc(start_date) - min_extra_rows
            if current_start_pos < 0:
                raise NoFurtherDataError.from_lookback_window(
                    initial_message="Insufficient data to compute Pipeline:",
                    first_date=all_dates[0],
                    lookback_start=start_date,
                    lookback_length=min_extra_rows,
                )
        except KeyError as exc:
            before, after = nearest_unequal_elements(all_dates, start_date)
            raise ValueError(
                "Pipeline start_date {start_date} is not in calendar.\n"
                "Latest date before start_date is {before}.\n"
                "Earliest date after start_date is {after}.".format(
                    start_date=start_date,
                    before=before,
                    after=after,
                )
            ) from exc

        # Our possible target dates are all the dates on or before the current
        # starting position.
        # TODO: Consider bounding this below by self.window_length
        candidates = all_dates[: current_start_pos + 1]

        # Choose the latest date in the candidates that is the start of a new
        # period at our frequency.
        choices = select_sampling_indices(candidates, self._frequency)

        # If we have choices, the last choice is the first date if the
        # period containing current_start_date.  Choose it.
        new_start_date = candidates[choices[-1]]

        # Add the difference between the new and old start dates to get the
        # number of rows for the new start_date.
        new_start_pos = all_dates.get_loc(new_start_date)
        assert new_start_pos <= current_start_pos, "Computed negative extra rows!"

        return min_extra_rows + (current_start_pos - new_start_pos)

    def _compute(self, inputs, dates, assets, mask):
        """
        Compute by delegating to self._wrapped_term._compute on sample dates.

        On non-sample dates, forward-fill from previously-computed samples.
        """
        to_sample = dates[select_sampling_indices(dates, self._frequency)]
        assert to_sample[0] == dates[0], (
                "Misaligned sampling dates in %s." % type(self).__name__
        )

        real_compute = self._wrapped_term._compute

        # Inputs will contain different kinds of values depending on whether or
        # not we're a windowed computation.

        # If we're windowed, then `inputs` is a list of iterators of ndarrays.
        # If we're not windowed, then `inputs` is just a list of ndarrays.
        # There are two things we care about doing with the input:
        #   1. Preparing an input to be passed to our wrapped term.
        #   2. Skipping an input if we're going to use an already-computed row.
        # We perform these actions differently based on the expected kind of
        # input, and we encapsulate these actions with closures so that we
        # don't clutter the code below with lots of branching.
        if self.windowed:
            # If we're windowed, inputs are stateful AdjustedArrays.  We don't
            # need to do any preparation before forwarding to real_compute, but
            # we need to call `next` on them if we want to skip an iteration.
            def prepare_inputs():
                return inputs

            def skip_this_input():
                for w in inputs:
                    next(w)

        else:
            # If we're not windowed, inputs are just ndarrays.  We need to
            # slice out a single row when forwarding to real_compute, but we
            # don't need to do anything to skip an input.
            def prepare_inputs():
                # i is the loop iteration variable below.
                return [a[[i]] for a in inputs]

            def skip_this_input():
                pass

        results = []
        samples = iter(to_sample)
        next_sample = next(samples)
        for i, compute_date in enumerate(dates):
            if next_sample == compute_date:
                results.append(
                    real_compute(
                        prepare_inputs(),
                        dates[i: i + 1],
                        assets,
                        mask[i: i + 1],
                    )
                )
                try:
                    next_sample = next(samples)
                except StopIteration:
                    # No more samples to take. Set next_sample to Nat, which
                    # compares False with any other datetime.
                    next_sample = pd_NaT
            else:
                skip_this_input()
                # Copy results from previous sample period.
                results.append(results[-1])

        # We should have exhausted our sample dates.
        try:
            next_sample = next(samples)
        except StopIteration:
            pass
        else:
            raise AssertionError("Unconsumed sample date: %s" % next_sample)

        # Concatenate stored results.
        return vstack(results)

    @staticmethod
    def _universal_mixin_type():
        return DownsampledMixin

    @staticmethod
    def _universal_mixin_specialization_name(principal_type):
        return "Downsampled" + principal_type.__name__


class SliceMixin(UniversalMixin):
    """Universal mixin for taking columnar slices of terms.

    Parameters
    ----------
    term : ziplime.pipeline.Term
        The term from which to extract a column of data.
    asset : ziplime.assets.Asset
        The asset corresponding to the column of `term` to be extracted.

    Notes
    -----
    Users should rarely construct instances of `Slice` directly. Instead, they
    should construct instances via indexing, e.g. `MyFactor()[Asset(24)]`.
    """

    def __new__(cls, term, asset):
        return super(SliceMixin, cls).__new__(
            cls,
            asset=asset,
            inputs=[term],
            window_length=0,
            mask=term.mask,
            dtype=term.dtype,
            missing_value=term.missing_value,
            window_safe=term.window_safe,
            ndim=1,
        )

    def __repr__(self):
        return "{parent_term}[{asset}]".format(
            parent_term=self.inputs[0].recursive_repr(),
            asset=self._asset,
        )

    def _init(self, asset, *args, **kwargs):
        self._asset = asset
        return super(SliceMixin, self)._init(*args, **kwargs)

    @classmethod
    def _static_identity(cls, asset, *args, **kwargs):
        return (
            super(SliceMixin, cls)._static_identity(*args, **kwargs),
            asset,
        )

    def _compute(self, windows, dates, assets, mask):
        asset = self._asset
        asset_column = searchsorted(assets.values, asset.sid)
        if assets[asset_column] != asset.sid:
            raise NonExistentAssetInTimeFrame(
                asset=asset,
                start_date=dates[0],
                end_date=dates[-1],
            )

        # Return a 2D array with one column rather than a 1D array of the
        # column.
        return windows[0][:, [asset_column]]

    @property
    def asset(self):
        """Get the asset whose data is selected by this slice."""
        return self._asset

    @staticmethod
    def _universal_mixin_type():
        return SliceMixin

    @staticmethod
    def _universal_mixin_specialization_name(principal_type):
        return principal_type.__name__ + "Slice"


class IfElseMixin(UniversalMixin):
    """Universal mixin for types returned by Filter.if_else."""

    window_length = 0

    def __new__(cls, condition: bool_dtype, if_true, if_false):
        return super(IfElseMixin, cls).__new__(
            cls,
            inputs=[condition, if_true, if_false],
            dtype=if_true.dtype,
            ndim=if_true.ndim,
            missing_value=if_true.missing_value,
            window_safe=all(
                (
                    condition.window_safe,
                    if_true.window_safe,
                    if_false.window_safe,
                )
            ),
            outputs=if_true.outputs,
        )

    def _compute(self, inputs, assets, dates, mask):
        if self.dtype == object:
            return labelarray_where(inputs[0], inputs[1], inputs[2])
        return where(inputs[0], inputs[1], inputs[2])

    @staticmethod
    def _universal_mixin_type():
        return IfElseMixin

    @staticmethod
    def _universal_mixin_specialization_name(principal_type):
        return "IfElse" + principal_type.__name__


class ConstantMixin(StandardOutputs, UniversalMixin):
    """Universal mixin for terms that produce a known constant value."""

    window_length = 0
    inputs = ()
    params = ("const",)

    def _compute(self, inputs, assets, dates, mask):
        constant = self.params["const"]
        out = full(mask.shape, constant, dtype=self.dtype)
        if self.dtype == object:
            return LabelArray(
                out,
                categories=[constant],
                missing_value=self.missing_value,
            )
        return out

    @staticmethod
    def _universal_mixin_type():
        return ConstantMixin

    @staticmethod
    def _universal_mixin_specialization_name(principal_type):
        return "Constant" + principal_type.__name__



from ziplime.finance.controls.trading_control import TradingControl


class MaxPositionSize(TradingControl):
    """TradingControl representing a limit on the maximum position size that can
    be held by an algo for a given asset.
    """

    def __init__(self, on_error, asset=None, max_shares=None, max_notional=None):
        super(MaxPositionSize, self).__init__(
            on_error, asset=asset, max_shares=max_shares, max_notional=max_notional
        )
        self.asset = asset
        self.max_shares = max_shares
        self.max_notional = max_notional

        if max_shares is None and max_notional is None:
            raise ValueError("Must supply at least one of max_shares and max_notional")

        if max_shares and max_shares < 0:
            raise ValueError("max_shares cannot be negative.")

        if max_notional and max_notional < 0:
            raise ValueError("max_notional must be positive.")

    def validate(self, asset, amount, portfolio, algo_datetime, algo_current_data):
        """Fail if the given order would cause the magnitude of our position to be
        greater in shares than self.max_shares or greater in dollar value than
        self.max_notional.
        """

        if self.asset is not None and self.asset != asset:
            return

        current_share_count = portfolio.positions[asset].amount
        shares_post_order = current_share_count + amount

        too_many_shares = (
                self.max_shares is not None and abs(shares_post_order) > self.max_shares
        )
        if too_many_shares:
            self.handle_violation(asset, amount, algo_datetime)

        current_price = algo_current_data.current(asset, "price")
        value_post_order = shares_post_order * current_price

        too_much_value = (
                self.max_notional is not None and abs(value_post_order) > self.max_notional
        )

        if too_much_value:
            self.handle_violation(asset, amount, algo_datetime)



from abc import ABC, abstractmethod

from ziplime.assets.entities.asset import Asset
from ziplime.finance.cancel_policy import NeverCancel

from ziplime.finance.commission import CommissionModel
from ziplime.finance.domain.order import Order
from ziplime.finance.domain.transaction import Transaction
from ziplime.domain.bar_data import BarData


class Blotter(ABC):
    def __init__(self, cancel_policy=None):
        self.cancel_policy = cancel_policy if cancel_policy else NeverCancel()

    @abstractmethod
    def save_order(self, order: Order) -> None: ...

    @abstractmethod
    def order_rejected(self, order: Order) -> None: ...

    @abstractmethod
    def order_held(self, order: Order) -> None: ...

    @abstractmethod
    def order_cancelled(self, order: Order) -> None: ...

    @abstractmethod
    def get_order_by_id(self, order_id: str, exchange_name: str) -> Order | None: ...

    @abstractmethod
    def get_open_orders_by_asset(self, asset: Asset, exchange_name: str) -> dict[str, Order] | None: ...

    @abstractmethod
    def get_open_orders(self,exchange_name: str) -> dict[Asset, dict[str, Order]]: ...

    @abstractmethod
    def get_all_assets_in_open_orders(self) -> list[Asset]: ...


    @abstractmethod
    def cancel_all_orders_for_asset(self, asset: Asset, exchange_name:str, relay_status: bool = True) -> None: ...

    def batch_order(self, order_arg_lists):
        """Place a batch of orders.

        Parameters
        ----------
        order_arg_lists : iterable[tuple]
            Tuples of args that `order` expects.

        Returns
        -------
        order_ids : list[str or None]
            The unique identifier (or None) for each of the orders placed
            (or not placed).

        Notes
        -----
        This is required for `Blotter` subclasses to be able to place a batch
        of orders, instead of being passed the order requests one at a time.
        """

        return [self.order(*order_args) for order_args in order_arg_lists]

    @abstractmethod
    def execute_cancel_policy(self, event):
        raise NotImplementedError("execute_cancel_policy")

    @abstractmethod
    def process_splits(self, splits: list[tuple[Asset, float]]) -> None:
        """
        Processes a list of splits by modifying any open orders as needed.

        Parameters
        ----------
        splits: list
            A list of splits.  Each split is a tuple of (asset, ratio).

        Returns
        -------
        None
        """

        raise NotImplementedError("process_splits")

    @abstractmethod
    def prune_orders(self, closed_orders) -> None:
        """
        Removes all given orders from the blotter's open_orders list.

        Parameters
        ----------
        closed_orders: iterable of orders that are closed.

        Returns
        -------
        None
        """

        raise NotImplementedError("prune_orders")



from abc import abstractmethod
from typing import Any

from ziplime.data.domain.data_bundle import DataBundle
from ziplime.data.services.bundle_storage import BundleStorage


class BundleRegistry:
    def __init__(self):
        pass

    async def list_bundles(self) -> list[dict[str, Any]]: ...

    async def list_bundles_by_name(self, bundle_name: str) -> list[dict[str, Any]]: ...

    async def load_bundle_metadata(self, bundle_name: str, bundle_version: str | None) -> dict[str, Any] | None: ...

    async def delete_bundle(self): ...

    @abstractmethod
    async def persist_metadata(self, data_bundle: DataBundle, metadata: dict[str, Any]): ...

    @abstractmethod
    async def get_bundle_metadata(self, data_bundle: DataBundle, bundle_storage: BundleStorage) -> dict[str, Any]: ...

    async def register_bundle(self, data_bundle: DataBundle, bundle_storage: BundleStorage):
        metadata = await self.get_bundle_metadata(data_bundle=data_bundle, bundle_storage=bundle_storage)
        await self.persist_metadata(data_bundle=data_bundle, metadata=metadata)



from .classifiers import Classifier, CustomClassifier
from .domain import Domain
from .factors import Factor, CustomFactor
from .filters import Filter, CustomFilter
from .term import Term, LoadableTerm, ComputableTerm
from .graph import ExecutionPlan, TermGraph

# NOTE: this needs to come after the import of `graph`, or else we get circular
# dependencies.
from .engine import SimplePipelineEngine
from .pipeline import Pipeline

__all__ = (
    "Classifier",
    "CustomFactor",
    "CustomFilter",
    "CustomClassifier",
    "Domain",
    "ExecutionPlan",
    "Factor",
    "Filter",
    "LoadableTerm",
    "ComputableTerm",
    "Pipeline",
    "SimplePipelineEngine",
    "Term",
    "TermGraph",
)
